/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include <math.h>
#include <stdio.h>
#include <zlib.h>
#include <mem.h>

#include "backdrop.h"

#include "bitmap.h"
#include "screen.h"

#define PNG_Signature       "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"

#define PNG_CN_IHDR 0x49484452L     /* Chunk names */
#define PNG_CN_PLTE 0x504C5445L
#define PNG_CN_IDAT 0x49444154L
#define PNG_CN_IEND 0x49454E44L
#define PNG_CN_gAMA 0x67414D41L
#define PNG_CN_sBIT 0x73424954L
#define PNG_CN_cHRM 0x6348524DL
#define PNG_CN_tRNS 0x74524E53L
#define PNG_CN_bKGD 0x624B4744L
#define PNG_CN_hIST 0x68495354L
#define PNG_CN_tEXt 0x74455874L
#define PNG_CN_zTXt 0x7A545874L
#define PNG_CN_pHYs 0x70485973L
#define PNG_CN_oFFs 0x6F464673L
#define PNG_CN_tIME 0x74494D45L
#define PNG_CN_sCAL 0x7343414CL

#define PNG_PF_None     0   /* Prediction filters */
#define PNG_PF_Sub      1
#define PNG_PF_Up       2
#define PNG_PF_Average  3
#define PNG_PF_Paeth    4

/* PNG support */
struct png_info {
	uint32 width, height;
	uint32 xoffset, yoffset;
	uint32 xres, yres;
	double xscale, yscale;
	double source_gamma;
	uint32 chromaticities[8];
	uint32 resolution_unit, offset_unit, scale_unit;
	uint8 bit_depth;
	uint32 significant_bits[4];
	uint32 background_color[4];
	uint8 color_type;
	uint8 compression_method;
	uint8 filter_method;
	uint8 interlace_method;
	uint32 num_palette;
	uint8 *palette;
	uint32 num_trans;
	uint8 *trans;
	uint8 *image;

	/* The rest is private and should not be used
	* by the public functions
	*/
	uint8 bpp;
	uint32 rowbytes;
	uint8 *zimage;
	uint32 zlength;
	uint8 *fimage;
};

extern char build_version[];

/* convert_uint is here so we don't have to deal with byte-ordering issues */
static uint32 convert_from_network_order (uint8 *v)
{
	uint32 i;
	i = (v[0]<<24) | (v[1]<<16) | (v[2]<<8) | (v[3]);
	return i;
}

static int png_unfilter(struct png_info *p, int (*fast_exit_handler)(void))
{
	int i, j, bpp, filter;
	int32 prediction, pA, pB, pC, dA, dB, dC;
	uint8 *src, *dst;

	if((p->image = (uint8 *)malloc (p->height*p->rowbytes))==NULL)
	{
/*              if (errorlog)
			fprintf(errorlog,"Out of memory\n");
*/
		free (p->fimage);
		return 0;
	}

	src = p->fimage;
	dst = p->image;
	bpp = p->bpp;

	for (i=0; i<p->height; i++)
	{
		/* fast exit */
		if (fast_exit_handler!=0 && fast_exit_handler()!=0) {
			free (p->fimage);
			return 0;
		}

		filter = *src++;
		if (!filter)
		{
			memcpy (dst, src, p->rowbytes);
			src += p->rowbytes;
			dst += p->rowbytes;
		}
		else
			for (j=0; j<p->rowbytes; j++)
			{
				pA = (j<bpp) ? 0: *(dst - bpp);
				pB = (i<1) ? 0: *(dst - p->rowbytes);
				pC = ((j<bpp)||(i<1)) ? 0: *(dst - p->rowbytes - bpp);

				switch (filter)
				{
				case PNG_PF_Sub:
					prediction = pA;
					break;
				case PNG_PF_Up:
					prediction = pB;
					break;
				case PNG_PF_Average:
					prediction = ((pA + pB) / 2);
					break;
				case PNG_PF_Paeth:
					prediction = pA + pB - pC;
					dA = abs(prediction - pA);
					dB = abs(prediction - pB);
					dC = abs(prediction - pC);
					if (dA <= dB && dA <= dC) prediction = pA;
					else if (dB <= dC) prediction = pB;
					else prediction = pC;
					break;
				default:
/*                                      if (errorlog)
						fprintf(errorlog,"Unknown filter type %i\n",filter);
*/
					prediction = 0;
				}
				*dst = 0xff & (*src + prediction);
				dst++; src++;
			}
	}

	free (p->fimage);
	return 1;
}

static int png_verify_signature (FILE *fp)
{
	int8 signature[8];

	if (fread(signature, 8, 1, fp) != 1)
	{
/*              if (errorlog)
			fprintf(errorlog,"Unable to read PNG signature (EOF)\n");
*/
		return 0;
	}

	if (memcmp(signature, PNG_Signature, 8))
	{
/*              if (errorlog)
			fprintf(errorlog,"PNG signature mismatch found: %s expected: %s\n",
				signature,PNG_Signature);
*/
		return 0;
	}
	return 1;
}

static int png_inflate_image (struct png_info *p,int (*fast_exit_handler)(void))
{
	unsigned long fbuff_size;

	(void)fast_exit_handler;

	fbuff_size = p->height * (p->rowbytes + 1);

	if((p->fimage = (uint8 *)malloc (fbuff_size))==NULL)
	{
/*              if (errorlog)
			fprintf(errorlog,"Out of memory\n");
*/
		free (p->zimage);
		return 0;
	}

	if (uncompress(p->fimage, &fbuff_size, p->zimage, p->zlength) != Z_OK)
	{
/*              if (errorlog)
			fprintf(errorlog,"Error while inflating image\n");
*/
		return 0;
	}

	free (p->zimage);
	return 1;
}

static int png_read_file(FILE *fp, struct png_info *p,int (*fast_exit_handler)(void))
{
	/* translates color_type to bytes per pixel */
	const int samples[] = {1, 0, 3, 1, 2, 0, 4};

	int i;
	uint32 chunk_length, chunk_type=0, chunk_crc, crc;
	uint32 tot_idat=0, num_idat=0, l_idat[256];
	uint8 *chunk_data, *zimage, *idat[256], *temp;
	uint8 str_chunk_type[5], v[4];

	p->num_palette = 0;
	p->num_trans = 0;
	p->trans = NULL;
	p->palette = NULL;
	p->image = NULL;

	if (png_verify_signature(fp)==0)
		return 0;

	while (chunk_type != PNG_CN_IEND)
	{
		if (fread(v, 4, 1, fp) != 1) {
/*                      if (errorlog)
				fprintf(errorlog,"Unexpected EOF in PNG\n"); */
		}
		chunk_length=convert_from_network_order(v);

		if (fread(str_chunk_type, 4, 1, fp) != 1) {
/*                      if (errorlog)
				fprintf(errorlog,"Unexpected EOF in PNG file\n"); */
		}

		str_chunk_type[4]=0; /* terminate string */

		crc=crc32(0,str_chunk_type, 4);
		chunk_type = convert_from_network_order(str_chunk_type);

		if (chunk_length)
		{
			if ((chunk_data = (uint8 *)malloc(chunk_length+1))==NULL)
			{
/*                              if (errorlog)
					fprintf(errorlog,"Out of memory\n"); */
				return 0;
			}
			if (fread (chunk_data, chunk_length, 1, fp) != 1)
			{
/*                              if (errorlog)
					fprintf(errorlog,"Unexpected EOF in PNG file\n");
*/
				free(chunk_data);
				return 0;
			}

			crc=crc32(crc,chunk_data, chunk_length);
		}
		else
			chunk_data = NULL;

		if (fread(v, 4, 1, fp) != 1) {
/*                      if (errorlog)
				fprintf(errorlog,"Unexpected EOF in PNG\n");
*/
		}
		chunk_crc=convert_from_network_order(v);

		if (crc != chunk_crc)
		{
/*                      if (errorlog)
			{
				fprintf(errorlog,"CRC check failed while reading PNG chunk %s\n",
					str_chunk_type);
				fprintf(errorlog,"Found: %08X  Expected: %08X\n",crc,chunk_crc);
			}
*/
			return 0;
		}

/*              if (errorlog)
			fprintf(errorlog,"Reading PNG chunk %s\n", str_chunk_type);
*/

		switch (chunk_type)
		{
		case PNG_CN_IHDR:
			p->width = convert_from_network_order(chunk_data);
			p->height = convert_from_network_order(chunk_data+4);
			p->bit_depth = *(chunk_data+8);
			p->color_type = *(chunk_data+9);
			p->compression_method = *(chunk_data+10);
			p->filter_method = *(chunk_data+11);
			p->interlace_method = *(chunk_data+12);
			free (chunk_data);

/*                      if (errorlog)
			{
				fprintf(errorlog,"PNG IHDR information:\n");
				fprintf(errorlog,"Width: %i, Height: %i\n", p->width, p->height);
				fprintf(errorlog,"Bit depth %i, color type: %i\n", p->bit_depth, p->color_type);
				fprintf(errorlog,"Compression method: %i, filter: %i, interlace: %i\n",
					p->compression_method, p->filter_method, p->interlace_method);
			}
*/
			break;

		case PNG_CN_PLTE:
			p->num_palette=chunk_length/3;
			p->palette=chunk_data;
/*                      if (errorlog)
				fprintf(errorlog,"%i palette entries\n", p->num_palette);
*/
			break;

		case PNG_CN_tRNS:
			p->num_trans=chunk_length;
			p->trans=chunk_data;
/*                      if (errorlog)
				fprintf(errorlog,"%i transparent palette entries\n", p->num_trans);
*/
			break;

		case PNG_CN_IDAT:
			idat[num_idat]=chunk_data;
			l_idat[num_idat]=chunk_length;
			num_idat++;
			tot_idat += chunk_length;
			break;

		case PNG_CN_tEXt:
/*                      if (errorlog)
			{
				uint8 *text=chunk_data;

				while(*text++);
				chunk_data[chunk_length]=0;
				fprintf(errorlog, "Keyword: %s\n", chunk_data);
				fprintf(errorlog, "Text: %s\n", text);
			}
*/
			free(chunk_data);
			break;

		case PNG_CN_tIME:
/*                      if (errorlog)
			{
				uint8 *t=chunk_data;
				fprintf(errorlog, "Image last-modification time: %i/%i/%i (%i:%i:%i) GMT\n",
					((short)(*t) << 8)+ (short)(*(t+1)), *(t+2), *(t+3), *(t+4), *(t+5), *(t+6));
			}
*/

			free(chunk_data);
			break;

		case PNG_CN_gAMA:
			p->source_gamma  = convert_from_network_order(chunk_data)/100000.0;
/*                      if (errorlog)
				fprintf(errorlog, "Source gamma: %f\n",p->source_gamma);
*/

			free(chunk_data);
			break;

		case PNG_CN_pHYs:
			p->xres = convert_from_network_order(chunk_data);
			p->yres = convert_from_network_order(chunk_data+4);
			p->resolution_unit = *(chunk_data+8);
/*                      if (errorlog)
			{
				fprintf(errorlog, "Pixel per unit, X axis: %i\n",p->xres);
				fprintf(errorlog, "Pixel per unit, Y axis: %i\n",p->yres);
				if (p->resolution_unit)
					fprintf(errorlog, "Unit is meter\n");
				else
					fprintf(errorlog, "Unit is unknown\n");
			}
*/
			free(chunk_data);
			break;

		case PNG_CN_IEND:
			break;

		default:
/*                      if (errorlog)
			{
				if (chunk_type & 0x40000000)
					fprintf (errorlog, "Ignoring critical chunk name %s!\n",str_chunk_type);
				else
					fprintf (errorlog, "Ignoring uncritical chunk name %s\n",str_chunk_type);
			}
*/
			if (chunk_data)
				free(chunk_data);
			break;
		}
	}
	if ((zimage = (uint8 *)malloc(tot_idat))==NULL)
	{
/*              if (errorlog)
			fprintf(errorlog,"Out of memory\n");
*/
		return 0;
	}

	p->zlength = tot_idat;

	/* combine idat chunks to compressed image data */
	temp = zimage;
	for (i=0; i<num_idat; i++)
	{
		memcpy (temp, idat[i], l_idat[i]);
		free (idat[i]);
		temp += l_idat[i];
	}
	p->zimage=zimage;
	p->bpp = (samples[p->color_type] * p->bit_depth) / 8;
	p->rowbytes = ceil((p->width * p->bit_depth * samples[p->color_type]) / 8.0);

	if (png_inflate_image(p,fast_exit_handler)==0)
		return 0;

	if(png_unfilter (p,fast_exit_handler)==0)
		return 0;

	return 1;
}

/*  Expands a p->image from p->bit_depth to 8 bit */
static int png_expand_buffer_8bit (struct png_info *p)
{
	int i,j, k;
	uint8 *inp, *outp, *outbuf;

	if (p->bit_depth < 8)
	{
		if ((outbuf = (uint8 *)malloc(p->width*p->height))==NULL)
		{
/*                      if (errorlog)
				fprintf(errorlog,"Out of memory\n"); */
			return 0;
		}

		inp = p->image;
		outp = outbuf;

		for (i = 0; i < p->height; i++)
		{
			for(j = 0; j < p->width / ( 8 / p->bit_depth); j++)
			{
				for (k = 8 / p->bit_depth-1; k >= 0; k--)
					*outp++ = (*inp >> k * p->bit_depth) & (0xff >> (8 - p->bit_depth));
				inp++;
			}
			if (p->width % (8 / p->bit_depth))
			{
				for (k = p->width % (8 / p->bit_depth)-1; k >= 0; k--)
					*outp++ = (*inp >> k * p->bit_depth) & (0xff >> (8 - p->bit_depth));
				inp++;
			}
		}
		free (p->image);
		p->image = outbuf;
	}
	return 1;
}

static void png_free(struct png_info *p) {
	if (p) {
		free(p->palette);
		free(p->trans);
		free(p->image);
	}
}

struct bitmap* backdrop_load(char *filename, video_color* rgb, unsigned* rgb_max, int (*fast_exit_handler)(void)) {
	FILE* f;
	struct png_info png;
	struct bitmap* bitmap;
	unsigned depth;

	f = fopen(filename,"rb");
	if (!f) {
		goto out;
	}

	if (png_read_file(f,&png,fast_exit_handler)==0) {
		goto out_close;
	}

	if (png.interlace_method != 0) {
		goto out_png;
	}

	/* limitations */
	if (png.color_type == 3 && png.bit_depth <= 8) {
		if (png_expand_buffer_8bit(&png)==0) {
			goto out_png;
		}

		if (!png.palette) {
			goto out_png;
		}

		depth = 8;
	} else if (png.color_type == 2 && png.bit_depth == 8) {
		depth = 24;
	} else {
		goto out_png;
	}

	bitmap = bitmap_alloc(png.width, png.height, depth);
	if (!bitmap) {
		goto out_png;
	}

	if (depth == 8) {
		int i;
		int y;

		unsigned char* palette = (unsigned char*)png.palette;
		for(i=0;i<png.num_palette;++i) {
			rgb[i].red = palette[0];
			rgb[i].green = palette[1];
			rgb[i].blue = palette[2];
			palette += 3;
		}
		*rgb_max = png.num_palette;

		for(y=0;y<png.height;++y) {
			unsigned char* src_off = png.image + y * png.width;
			uint8* dst_off = bitmap_line(bitmap,y);
			memcpy(dst_off, src_off, png.width);
		}
	} else {
		int y;
		for(y=0;y<png.height;++y) {
			unsigned char* src_off = png.image + y * 3 * png.width;
			uint8* dst_off = bitmap_line(bitmap,y);
			memcpy(dst_off, src_off, 3*png.width);
		}
	}

	fclose(f);
	png_free(&png);
	return bitmap;
out_png:
	png_free(&png);
out_close:
	fclose(f);
out:
	return 0;
}
