/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "game.h"
#include "text.h"
#include "menu.h"

#include <unistd.h>

#include <strstream>
#include <iomanip>
#include <algorithm>



// 9 blue
// A green
// B cyan
// C red
// D magenta
// E yellow
// F white

// -------------------------------------------------------------------------
//

unsigned COLOR_HELP_NORMAL = 0xF0;
unsigned COLOR_HELP_TAG = 0xFC;
unsigned COLOR_CHOICE_TITLE = 0xFC;
unsigned COLOR_CHOICE_NORMAL = 0xF0;
unsigned COLOR_CHOICE_SELECT = 0xA0;
unsigned COLOR_MENU_NORMAL = 0xF0;
unsigned COLOR_MENU_HIDDEN = 0xF8;
unsigned COLOR_MENU_TAG = 0xFC;
unsigned COLOR_MENU_SELECT = 0xA0;
unsigned COLOR_MENU_HIDDEN_SELECT = 0xA8;
unsigned COLOR_MENU_TAG_SELECT = 0xAC;
unsigned COLOR_MENU_BAR = 0xF0;
unsigned COLOR_MENU_BAR_TAG = 0xFC;
unsigned COLOR_MENU_BAR_HIDDEN = 0xF8;
unsigned COLOR_MENU_GRID = 0xFC;
unsigned COLOR_MENU_BACKDROP = 0x80;

static struct {
	unsigned* var;
	const char* name;
} COLOR_TAB[] = {
{ &COLOR_HELP_NORMAL, "help" },
{ &COLOR_HELP_TAG, "help_tag" },
{ &COLOR_CHOICE_TITLE, "submenu_bar" },
{ &COLOR_CHOICE_NORMAL, "submenu_item" },
{ &COLOR_CHOICE_SELECT, "submenu_item_select" },
{ &COLOR_MENU_NORMAL, "menu_item" },
{ &COLOR_MENU_HIDDEN, "menu_hidden" },
{ &COLOR_MENU_TAG, "menu_tag" },
{ &COLOR_MENU_SELECT, "menu_item_select" },
{ &COLOR_MENU_HIDDEN_SELECT, "menu_hidden_select" },
{ &COLOR_MENU_TAG_SELECT, "menu_tag_select" },
{ &COLOR_MENU_BAR, "bar" },
{ &COLOR_MENU_BAR_TAG, "bar_tag" },
{ &COLOR_MENU_BAR_HIDDEN, "bar_hidden" },
{ &COLOR_MENU_GRID, "grid" },
{ &COLOR_MENU_BACKDROP, "backdrop" },
{ 0,0 }
};

static char* COLOR[] = {
"black",
"blue",
"green",
"cyan",
"red",
"magenta",
"brown",
"lightgray",
"gray",
"lightblue",
"lightgreen",
"lightcyan",
"lightred",
"lightmagenta",
"yellow",
"white"
};

static unsigned string2color(const string& s) {
	for(unsigned i=0;i<16;++i)
		if (COLOR[i] == s)
			return i;
	return 0;
}

static const string color2string(unsigned c) {
	if (c < 16)
		return COLOR[c];
	else
		return COLOR[0];
}

bool color_in(const string& s) {
	string sname;
	string sarg0;
	string sarg1;
	unsigned i = 0;

	while (i < s.length() && !isspace(s[i])) {
		sname += s[i];
		++i;
	}
	
	while (i < s.length() && isspace(s[i]))
		++i;

	while (i < s.length() && !isspace(s[i])) {
		sarg0 += s[i];
		++i;
	}
	
	while (i < s.length() && isspace(s[i]))
		++i;

	while (i < s.length() && !isspace(s[i])) {
		sarg1 += s[i];
		++i;
	}

	while (i < s.length() && isspace(s[i]))
		++i;

	if (i != s.length())
		return false;

	for(i=0;COLOR_TAB[i].name;++i) {
		if (COLOR_TAB[i].name == sname)
			break;
	}

	if (!COLOR_TAB[i].name)
		return false;

	*COLOR_TAB[i].var = string2color(sarg0) | string2color(sarg1) << 4;

	return true;
}

void color_out_all(ostream& os) {
	unsigned i;
	for(i=0;COLOR_TAB[i].name;++i) {
		os << "color " << COLOR_TAB[i].name << " ";
		os << color2string( (*COLOR_TAB[i].var) & 0xF);
		os << " ";
		os << color2string( ((*COLOR_TAB[i].var) >> 4) & 0xF);
		os << endl;
	}
}

// -------------------------------------------------------------------------
// Draw

int draw_string(int x, int y, int cx, const string& s, int color) {
	if (cx>0) {
		if (s.length() >= cx) {
			text_put(x,y,string(s,0,cx),color);
		} else {
			text_put(x,y,s,color);
			text_fill(x+s.length()*text_font_dx_get(),y,cx-s.length(),' ',color);
		}
		return cx;
	} else {
		return 0;
	}
}

int draw_string(int x, int y, const string& s, int color) {
	return draw_string(x,y,s.length(),s,color);
}

int draw_char(int x, int y, int cx, char c, int color) {
	if (cx>0) {
		text_fill(x,y,cx,c,color);
		return cx;
	} else {
		return 0;
	}
}

void draw_string_special(int x, int y, int len, const string& s, int c0, int c1, int c2, bool print_in) {
	bool in = false;
	for(int i=0;i<len;++i) {
		if (i==0) {
			draw_char(x,y,1,s[i],c0);
		} else {
			if (s[i]=='(' || s[i]=='[')
				in = true;
			if (!in || print_in) {
				draw_char(x,y,1,s[i], in ? c1 : c2);
			} else {
				draw_char(x,y,1,' ', c2);
			}
			if (s[i]==')' || s[i]==']')
				in = false;
		}
		x += text_font_dx_get();
	}
}

// ------------------------------------------------------------------------
// Choice menu

struct choice {
	choice(const string& Adesc);
	choice(const string& Adesc, bool Abistate);
	choice(const string& Adesc, tristate_t Atristate);

	string desc;
	int state;
	bool bistate;
	tristate_t tristate;
};

choice::choice(const string& Adesc) :
	desc(Adesc), state(1) {
}

choice::choice(const string& Adesc, bool Abistate) :
	desc(Adesc),state(2),bistate(Abistate) {
}

choice::choice(const string& Adesc, tristate_t Atristate) :
	desc(Adesc),state(3),tristate(Atristate) {
}

typedef list<choice> choice_container;

void draw_choice_window(const string& title, int x, int y, int cx, choice_container& ch, choice_container::iterator& pos) {
	draw_string(x,y,cx,title, COLOR_CHOICE_TITLE);
	cx -= 6;
	y += text_font_dy_get();
	for(choice_container::iterator i = ch.begin();i!=ch.end();++i) {
		int color;
		if (i==pos) {
			color = COLOR_CHOICE_SELECT;
		} else {
			color = COLOR_CHOICE_NORMAL;
		}
		const char* sstate;
		if (i->state==1) {
			if (i==pos) {
				sstate = " * ";
			} else {
				sstate = "   ";
			}
		} else if (i->state==2) {
			if (i->bistate) {
				sstate = " + ";
			} else {
				sstate = " - ";
			}
		} else {
			if (i->tristate == exclude) {
				sstate = " - ";
			} else if (i->tristate == exclude_not) {
				sstate = " * ";
			} else {
				sstate = " + ";
			}
		}
		draw_string(x,y,sstate, color);
		draw_string(x+3*text_font_dx_get(),y,cx,i->desc, color);
		draw_string(x+(3+cx)*text_font_dx_get(),y,"   ", color);
		y += text_font_dy_get();
	}
}

int run_choice(const string& title, int x, int y, int dx, choice_container& ch, choice_container::iterator& pos) {
	int key = TEXT_KEY_ESC;
	int done = 0;
	int dy = (ch.size()+1)*text_font_dy_get();

	int border = text_font_dx_get()/2;
	text_box(x-border,y-border,dx+2*border,dy+border*2,COLOR_CHOICE_NORMAL);
	text_clear(x-border+1,y-border+1,dx+2*border-2,dy+border*2-2,COLOR_CHOICE_NORMAL >> 4);

	while (!done) {
		draw_choice_window(title,x,y,dx / text_font_dx_get(),ch,pos);
		key = text_getkey();

		switch (key) {
			case TEXT_KEY_DEL :
				for(choice_container::iterator i=ch.begin();i!=ch.end();++i) {
					switch (i->state) {
						case 2: i->bistate = false; break;
						case 3: i->tristate = exclude; break;
					}
				}
			break;
			case TEXT_KEY_INS :
				for(choice_container::iterator i=ch.begin();i!=ch.end();++i) {
					switch (i->state) {
						case 2: i->bistate = true; break;
						case 3: i->tristate = include; break;
					}
				}
			break;
			case TEXT_KEY_HOME :
				pos = ch.begin();;
				break;
			case TEXT_KEY_END :
				if (pos != ch.end()) {
					do {
						++pos;
					} while (pos != ch.end());
					--pos;
				}
				break;
			case TEXT_KEY_UP :
				if (pos != ch.begin())
					--pos;
				break;
			case TEXT_KEY_DOWN :
				if (pos != ch.end()) {
					if (++pos == ch.end())
						--pos;
				}
				break;
			case TEXT_KEY_SPACE :
				switch (pos->state) {
					case 2: pos->bistate = !pos->bistate; break;
					case 3:
						switch (pos->tristate) {
							case include : pos->tristate = exclude; break;
							case exclude : pos->tristate = exclude_not; break;
							case exclude_not : pos->tristate = include; break;
						}
				}
				break;
			case TEXT_KEY_ENTER :
			case TEXT_KEY_ESC :
				done = 1;
				break;
		}
	}

	return key;
}

// ------------------------------------------------------------------------
// menu draw

void draw_menu_game(const game_set& gar, const game* g, int x, int y, int dx, bool selected, bool ident_if_clone, merge_t merge, bool clone_listed) {
	string s;
	if (g)
		s = g->description_get();
	else
		s = "";

	int len = s.length();
	unsigned ident;
	if (ident_if_clone && g && g->parent_get())
		ident = 2*text_font_dx_get() + text_font_dx_get()/2;
	else
		ident = text_font_dx_get()/2;
	if (len*text_font_dx_get() + ident > dx)
		len = (dx - ident)/text_font_dx_get();

	int color;
	int color_first;
	int color_in;
	const game* gb;
 	if (!clone_listed && g)
 		gb = g->clone_best_get();
	else
		gb = g;
	if (gb && gb->is_good() && gar.is_tree_rom_of_present(gb->name_get(),merge)) {
		color = selected ? COLOR_MENU_SELECT : COLOR_MENU_NORMAL;
		color_first = selected ? COLOR_MENU_TAG_SELECT : COLOR_MENU_TAG;
		color_in = selected ? COLOR_MENU_HIDDEN_SELECT : COLOR_MENU_HIDDEN;
	} else {
		color = selected ? COLOR_MENU_HIDDEN_SELECT : COLOR_MENU_HIDDEN;
		color_first = color;
		color_in = color;
	}

	text_clear(x,y,ident,text_font_dy_get(),color >> 4);
	if (len)
		draw_string_special(x+ident,y,len,s,color_first,color_in,color,clone_listed);
	text_clear(x+ident+len*text_font_dx_get(),y,dx-ident-len*text_font_dx_get(), text_font_dy_get(),color >> 4);
}

void draw_tag_left(const string& s, int& xl, int& xr, int y, int sep, int color) {
	int len = sep + s.length();
	if (len < xr-xl) {
		draw_string(xl*text_font_dx_get(),y,s,color);
		xl += len;
	}
}

void draw_tag_right(const string& s, int& xl, int& xr, int y, int sep, int color) {
	int len = s.length();
	if (len+sep < xr-xl) {
		draw_string((xr-len)*text_font_dx_get(),y,s,color);
		xr -= len + sep;
	}
}

void draw_menu_bar(const game* g, int g1, int g2, int x, int y, int dx, bool clone_listed) {
	text_clear(x,y,dx,text_font_dy_get(),COLOR_MENU_BAR >> 4);

	int cx = dx / text_font_dx_get();
	int separator = cx > 40 ? 1 : 0;
	int in_separator =  cx > 40 ? 2 : 1;
	int xr = cx - separator + x;
	int xl = separator + x;

	if (g) {
		ostrstream os;
		os << setw(4) << setfill(' ') << g1 << "/" << setw(4) << setfill(' ') << g2 << ends;
		draw_tag_right(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
		os.freeze(0);
	}

	if (g) {
		ostrstream os;
		os << setw(3) << setfill('0') << (g->time_get()/3600) << ":" << setw(2) << setfill('0') << ((g->time_get() / 60) % 60) << ends;
		draw_tag_right(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
		os.freeze(0);
	}

	if (g) {
		ostrstream os;
		if (g->size_get() >= 10*1000*1000)
			os << setw(4) << g->size_get()/1000/1000 << "M"  << ends;
		else
			os << setw(4) << g->size_get()/1000 << "k"  << ends;
		draw_tag_right(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
		os.freeze(0);
	}

	if (g && g->sizex_get() && g->sizey_get()) {
		ostrstream os;
		os << g->sizex_get() << "x" << g->sizey_get() << ends;
		draw_tag_right(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
		os.freeze(0);
	}

	if (g && g->group_derived_get() != GROUP_UNDEFINED) {
		ostrstream os;
		os << g->group_derived_get() << ends;
		draw_tag_right(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
		os.freeze(0);
	}

	if (g && g->type_derived_get() != TYPE_UNDEFINED) {
		ostrstream os;
		os << g->type_derived_get() << ends;
		draw_tag_right(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
		os.freeze(0);
	}

	if (g) {
		ostrstream os;
		os << g->name_get() << ends;
		draw_tag_left(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
		os.freeze(0);
	}

	if (!clone_listed && g) {
		const game* gb = g->clone_best_get();
		if (gb != g) {
			ostrstream os;
			os << "[" << gb->name_get() << "]" << ends;
			draw_tag_left(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
			os.freeze(0);
		}
	}
}

void draw_menu_info(const game_set& gar, const game* g, int x, int y, int dx, merge_t merge, preview_t preview, bool clone_listed) {
	text_clear(x,y,dx,text_font_dy_get(),COLOR_MENU_BAR >> 4);

	int cx = dx / text_font_dx_get();
	int separator = cx > 40 ? 1 : 0;
	int in_separator =  cx > 40 ? 2 : 1;
	int xr = cx - separator + x;
	int xl = separator + x;

	if (g) {
		const game* gb;
	 	if (!clone_listed && g)
	 		gb = g->clone_best_get();
		else
			gb = g;

		if (!gar.is_tree_rom_of_present(gb->name_get(),merge))
			draw_tag_right("MISSING",xl,xr,y,in_separator,COLOR_MENU_BAR_TAG);
		else if (!gb->working_get())
			draw_tag_right("  ALPHA",xl,xr,y,in_separator,COLOR_MENU_BAR_TAG);
		else if (!gb->color_get())
			draw_tag_right("NOCOLOR",xl,xr,y,in_separator,COLOR_MENU_BAR_TAG);
		else if (!gb->sound_get())
			draw_tag_right("NOSOUND",xl,xr,y,in_separator,COLOR_MENU_BAR_TAG);
		else
			draw_tag_right("       ",xl,xr,y,in_separator,COLOR_MENU_BAR_TAG);

		draw_tag_left(g->manufacturer_get(),xl,xr,y,0,COLOR_MENU_BAR);

		if (g->year_get().length())
			draw_tag_left(", " + g->year_get(),xl,xr,y,0,COLOR_MENU_BAR);

		if (g->clone_get() > 0) {
			ostrstream os;
			os << ", " << g->clone_get() << " clones" << ends;
			draw_tag_left(os.str(),xl,xr,y,0,COLOR_MENU_BAR);
			os.freeze(0);
		}

		if (g->parent_get()) {
			ostrstream os;
			os << ", cloneof " << g->parent_get()->name_get() << ends;
			draw_tag_left(os.str(),xl,xr,y,in_separator,COLOR_MENU_BAR);
			os.freeze(0);
		}

		switch (preview) {
			case preview_flyers: draw_tag_right("flyers",xl,xr,y,in_separator,COLOR_MENU_BAR_HIDDEN); break;
			case preview_snap: break;
		}
	}
}

struct cell_t {
	int x; // position
	int y;
	int dx; // size
	int dy;
};

void draw_menu_window(const game_set& gar, const pgame_array& gc, struct cell_t* cell, int coln, int rown, int start, int pos, bool ident_if_clone, merge_t merge, bool clone_listed) {
	for(int r=0;r<rown;++r) {
		for(int c=0;c<coln;++c) {
			const game* g = start<gc.size() ? gc[start] : 0;
			draw_menu_game(gar,g,cell->x,cell->y+cell->dy-text_font_dy_get(),cell->dx, start == pos, ident_if_clone, merge, clone_listed );
			++start;
			++cell;
		}
	}
}

void draw_menu_scroll(int x, int y, int dx, int dy, int pos, int delta, int max) {
	if (max <= 1)
		return;

	if (pos >= max)
		pos = max - 1;
	if (pos + delta > max)
		delta = max - pos;

	int y0 = pos * (dy-1) / (max-1);
	int y1 = (pos+delta-1) * (dy-1) / (max-1);

	text_clear(x,y,dx,dy,COLOR_MENU_GRID >> 4);
	text_clear(x,y+y0,dx,y1-y0+1,COLOR_MENU_GRID);
}

// ------------------------------------------------------------------------
// Menu utility

void menu_pos(int pos, menu_state& rs, int pos_max, int base_max, int coln) {
	while (pos >= base_max)
		pos -= coln;
	while (pos < 0)
		pos += coln;
	if (pos >= rs.pos_base && pos < rs.pos_base + pos_max) {
		rs.pos_rel = pos - rs.pos_base;
	} else if (pos < rs.pos_base) {
		rs.pos_rel = pos % coln;
		rs.pos_base = pos - rs.pos_rel;
	} else {
		rs.pos_rel = pos % coln + pos_max - coln;
		rs.pos_base = pos - rs.pos_rel;
		if (rs.pos_base < 0) {
			rs.pos_base = 0;
			rs.pos_rel = pos;
		}
	}
}

void menu_center(int pos, menu_state& rs, int pos_max, int base_max, int coln) {
	while (pos >= base_max)
		pos -= coln;
	while (pos < 0)
		pos += coln;
	if (pos >= rs.pos_base && pos < rs.pos_base + pos_max) {
		rs.pos_rel = pos - rs.pos_base;
	} else {
		rs.pos_rel = pos % coln + (pos_max / coln) / 2 * coln;
		rs.pos_base = pos - rs.pos_rel;
		if (rs.pos_base < 0) {
			rs.pos_base = 0;
			rs.pos_rel = pos;
		}
	}
}

bool menu_fast_compare(const string& game, const string& fast) {
	unsigned igame = 0;
	unsigned ifast = 0;
	while (igame < game.length() && ifast < fast.length()) {
		while (igame < game.length() && !isalnum(game[igame]))
		++igame;
		if (igame < game.length()) {
			if (toupper(game[igame]) != toupper(fast[ifast]))
				return false;
			++igame;
			++ifast;
		}
	}
	return ifast == fast.length();
}

//--------------------------------------------------------------------------
// Backdrop

bool backdrop_find_single(string& path, const string& snap, const string& name) {
	unsigned i = 0;
	while (i<snap.length()) {
		int end = snap.find(';',i);
		if (end==string::npos) {
			path = string( snap, i );
			i = snap.size();
		} else {
			path = string( snap, i, end-i );
			i = end + 1;
		}
		if (path.length()) {
			if (path[path.length()-1]!='\\' && path[path.length()-1]!='/') {
				path += "/";
			}
		}

		path += name + ".png";

		// test read access
		if (access(path.c_str(),F_OK | R_OK)==0) {
			return true;
		}
	}
	return false;
}

bool backdrop_find_clone(string& path, string snap, const game* pgame) {
	if (pgame) {
		// search game
		if (backdrop_find_single(path,snap,pgame->name_get()))
			return true;
		// search clones
		for(pgame_container::const_iterator i=pgame->clone_bag_get().begin();i!=pgame->clone_bag_get().end();++i) {
			if (backdrop_find_clone(path,snap,*i))
				return true;
		}
	}
	return false;
}

bool backdrop_find_all(string& path, string snap, const game* pgame) {
	if (pgame) {
		// search clone
		if (backdrop_find_clone(path,snap,pgame))
			return true;
		// search root
		if (pgame->parent_get()) {
			const game* proot = pgame->parent_get();
			while (proot->parent_get())
				proot =	proot->parent_get();
			if (backdrop_find_clone(path,snap,proot))
				return true;
		}
	}
	return false;
}

string backdrop_cat(const string& A, const string& B) {
	string r;
	r = A;
	if (A.length() && A[A.length()-1]!=';' && B.length())
		r += ";" + B;
	return r;
}

bool backdrop_find_preview(string& path, const menu_state& rs, const game* pgame) {
	string search_path;
	switch (rs.preview) {
		case preview_snap : search_path = backdrop_cat(rs.snap,rs.flyers); break;
		case preview_flyers : search_path = backdrop_cat(rs.flyers,rs.snap); break;
	}
	return backdrop_find_all(path,search_path,pgame);
}

//--------------------------------------------------------------------------
// Menu run

static int run_menu_user(const game_set& gar, menu_state& rs, bool flipxy, pgame_array& gc, int game_good) {
	int coln; // number of columns
	int rown; // number of rows
	int colrown; // colums * rowns

	int win_x; // menu window
	int win_y;
	int win_dx;
	int win_dy;

	int backdrop_x; // backdrop window (if one)
	int backdrop_y;
	int backdrop_dx;
	int backdrop_dy;

	int bar_top_x; // top bar
	int bar_top_y;
	int bar_top_dx;
	int bar_top_dy;

	int bar_bottom_x; // bottom bar
	int bar_bottom_y;
	int bar_bottom_dx;
	int bar_bottom_dy;

	int bar_left_x; // left bar
	int bar_left_y;
	int bar_left_dx;
	int bar_left_dy;

	int bar_right_x; // Right bar
	int bar_right_y;
	int bar_right_dx;
	int bar_right_dy;

	int space_x; // space beetween tiles
	int space_y;

	bar_top_x = 0;
	bar_top_y = 0;
	bar_top_dy = text_font_dy_get();

	bar_bottom_x = 0;
	bar_bottom_dy =	text_font_dy_get();

	bar_left_x = 0;
	bar_left_y = bar_top_dy;
	bar_left_dx = text_font_dx_get()/2;

	bar_right_y = bar_top_dy;
	bar_right_dx = text_font_dx_get()/2;

	win_x = bar_left_dx;
	win_y = bar_top_dy;

	if (rs.mode == mode_tile_4x3 || rs.mode == mode_tile_3x2) {
		space_x = text_font_dx_get()/2;
		space_y = text_font_dx_get()/2;
		if (!flipxy) {
			if (rs.mode == mode_tile_4x3) {
				coln = 4;
				rown = 3;
			} else {
				coln = 3;
				rown = 2;
			}
		} else {
			if (rs.mode == mode_tile_4x3) {
				coln = 3;
				rown = 4;
			} else {
				coln = 2;
				rown = 3;
			}
		}

		win_dx = (text_dx_get() - bar_left_dx - bar_right_dx - (coln-1) * space_x) / coln * coln + (coln-1) * space_x;
		win_dy = (text_dy_get() - bar_top_dy - bar_bottom_dy - (rown-1) * space_y) / rown * rown + (rown-1) * space_y;

		backdrop_x = 0;
		backdrop_y = 0;
		backdrop_dx = 0;
		backdrop_dy = 0;

		bar_right_x = bar_left_dx + win_dx;
		bar_bottom_y = bar_top_dy + win_dy;
		bar_top_dx = bar_bottom_dx = bar_left_dx + bar_right_dx + win_dx;
		bar_left_dy = bar_right_dy = win_dy;
	} else if (rs.mode == mode_list) {
		if (!flipxy) {
			// horizontal
			space_x = 0;
			space_y = 0;

			coln = 1;
			win_dx = (text_dx_get() - bar_left_dx - bar_right_dx) / 3;
			win_dx -= win_dx % text_font_dx_get();

			rown = (text_dy_get() - bar_top_dy - bar_bottom_dy) / text_font_dy_get();
			win_dy = rown * text_font_dy_get();

			backdrop_x = win_x + win_dx;
			backdrop_y = win_y;
			backdrop_dx = text_dx_get() - win_dx - bar_left_dx - bar_right_dx;
			backdrop_dy = win_dy;

			bar_right_x = bar_left_dx + win_dx + backdrop_dx;
			bar_bottom_y = bar_top_dy + win_dy;
			bar_top_dx = bar_bottom_dx = bar_left_dx + bar_right_dx + win_dx + backdrop_dx;
			bar_left_dy = bar_right_dy = win_dy;
		} else {
			// vertical
			space_x = text_font_dx_get() / 2;
			space_y = 0;
			coln = text_dx_get()/(20*text_font_dx_get());
			if (coln < 2)
				coln = 2;
			win_dx = (text_dx_get() - bar_left_dx - bar_right_dx - (coln-1) * space_x) / coln * coln + (coln-1) * space_x;

			rown = ((text_dy_get() - bar_top_dy - bar_bottom_dy)/text_font_dy_get()) / 3;
			win_dy = rown * text_font_dy_get();

			backdrop_x = win_x;
			backdrop_y = win_y + win_dy;
			backdrop_dx = win_dx;
			backdrop_dy = text_dy_get() - win_dy - bar_top_dy - bar_bottom_dy;

			bar_right_x = bar_left_dx + win_dx;
			bar_bottom_y = bar_top_dy + win_dy + backdrop_dy;
			bar_top_dx = bar_bottom_dx = bar_left_dx + bar_right_dx + win_dx;
			bar_left_dy = bar_right_dy = win_dy + backdrop_dy;
		}
	} else
		return TEXT_KEY_ESC;

	colrown = coln*rown;

	struct cell_t* cell_map = new cell_t[colrown];

	int cell_dx = (win_dx - space_x * (coln-1)) / coln;
	int cell_dy = (win_dy - space_y * (rown-1)) / rown;

	for(int r=0;r<rown;++r) {
		for(int c=0;c<coln;++c) {
			struct cell_t* cell = cell_map + r*coln+c;
			cell->x = win_x + (cell_dx + space_x) * c;
			cell->y = win_y + (cell_dy + space_y) * r;
			cell->dx = cell_dx;
			cell->dy = cell_dy;
		}
	}

	if (rs.mode == mode_list) {
		text_backdrop_init(COLOR_MENU_BACKDROP, 1);
		text_backdrop_set(0,backdrop_x,backdrop_y,backdrop_dx,backdrop_dy);
	} else if (rs.mode == mode_tile_4x3 || rs.mode == mode_tile_3x2) {
		text_backdrop_init(COLOR_MENU_BACKDROP, colrown);
		for(int i=0;i<colrown;++i)
			text_backdrop_set(i,cell_map[i].x,cell_map[i].y,cell_map[i].dx,cell_map[i].dy-text_font_dy_get());
	}

	int pos_max = colrown;
	int base_max = gc.size();
	if (base_max % coln)
		base_max = base_max + coln - base_max % coln;

	if (rs.pos_base % coln) {
		rs.pos_rel += rs.pos_base % coln;
		rs.pos_base -= rs.pos_base % coln;
	}

	while (rs.pos_rel >= pos_max) {
		rs.pos_base += coln;
		rs.pos_rel -= coln;
	}

	if (rs.pos_base + rs.pos_rel >= base_max) {
		rs.pos_base = base_max - pos_max;
		if (rs.pos_base < 0) {
			rs.pos_base = 0;
		}
		if (rs.pos_base + rs.pos_rel >= base_max) {
			rs.pos_base = 0;
			rs.pos_rel = 0;
		}
	}

	if (rs.pgame) {
		unsigned i = 0;
		while (i<gc.size() && rs.pgame != gc[i])
			++i;
		if (i<gc.size())
			menu_pos(i,rs,pos_max,base_max,coln);
	} else if (rs.last_game.length()) {
		unsigned i = 0;
		while (i<gc.size() && rs.last_game != gc[i]->name_get())
			++i;
		if (i<gc.size())
			menu_pos(i,rs,pos_max,base_max,coln);
	}

	if (rs.pos_rel + rs.pos_base < gc.size()) {
		rs.pgame = gc[rs.pos_rel+rs.pos_base];
		rs.pclone = 0;
	} else {
		rs.pgame = 0;
		rs.pclone = 0;
	}

	bool done = false;
	int key = 0;
	int exit_count = 0;

	text_clear();

	while (!done) {

		if (space_x)
			for(unsigned i=0;i<coln-1;++i)
				text_clear(cell_map[i].x + cell_map[i].dx,win_y,space_x,win_dy,COLOR_MENU_GRID >> 4);
		if (space_y)
			for(unsigned i=0;i<rown-1;++i)
				text_clear(win_x,cell_map[i*coln].y + cell_map[i*coln].dy,win_dx,space_y,COLOR_MENU_GRID >> 4);

		draw_menu_window(gar,gc,cell_map,coln,rown,rs.pos_base, rs.pos_base + rs.pos_rel, rs.sort_mode == sort_by_root_name, rs.merge, rs.exclude_clone != exclude);
		draw_menu_bar(rs.pgame,game_good,gc.size(),bar_top_x,bar_top_y,bar_top_dx, rs.exclude_clone != exclude);
		draw_menu_info(gar,rs.pgame,bar_bottom_x,bar_bottom_y,bar_bottom_dx,rs.merge,rs.preview, rs.exclude_clone != exclude);
		draw_menu_scroll(bar_right_x,bar_right_y,bar_right_dx,bar_right_dy,rs.pos_base,pos_max,base_max);
		text_clear(bar_left_x,bar_left_y,bar_left_dx,bar_left_dy,COLOR_MENU_BAR >> 4);

		if (rs.mode == mode_list) {
			string backdrop;
			if (backdrop_find_preview(backdrop,rs,rs.pgame)) {
				text_backdrop_set(0, backdrop.c_str(), false );
			} else {
				text_backdrop_clear(0, false);
			}
			text_backdrop_update();
		} else if (rs.mode == mode_tile_4x3 || rs.mode == mode_tile_3x2) {
			for(int i=0;i<colrown;++i) {
				string backdrop;
				const game* pgame = rs.pos_base + i < gc.size() ? gc[rs.pos_base + i] : 0;
				if (backdrop_find_preview(backdrop,rs,pgame)) {
					text_backdrop_set(i, backdrop.c_str(), i == rs.pos_rel );
				} else {
					text_backdrop_clear(i, i == rs.pos_rel);
				}
			}
			text_backdrop_update();
		}

		key = text_getkey();

		string oldfast = rs.fast;
		rs.fast.erase();

		if (key != TEXT_KEY_ESC)
			exit_count = 0;

		switch (key) {
			case TEXT_KEY_SPACE :
				switch (rs.preview) {
					case preview_snap: rs.preview = preview_flyers; break;
					case preview_flyers: rs.preview = preview_snap; break;
				}
				break;
			case TEXT_KEY_HOME :
				menu_pos(0,rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_END :
				menu_pos(gc.size()-1,rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_PGUP :
				if (rs.pos_base >= pos_max) {
					rs.pos_base -= pos_max;
				} else if (rs.pos_base>0) {
					rs.pos_base = 0;
				} else {
					rs.pos_rel = 0;
				}
				break;
			case TEXT_KEY_PGDN :
				if (rs.pos_base + 2*pos_max <= base_max) {
					rs.pos_base += pos_max;
				} else if (rs.pos_base < base_max - pos_max) {
					rs.pos_base = base_max - pos_max;
				} else {
					rs.pos_rel = gc.size() - rs.pos_base - 1;
				}
				break;
			case TEXT_KEY_UP :
				menu_pos(rs.pos_base + rs.pos_rel - coln,rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_DOWN :
				menu_pos(rs.pos_base + rs.pos_rel + coln,rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_LEFT :
				menu_pos(rs.pos_base + rs.pos_rel -1,rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_RIGHT :
				menu_pos(rs.pos_base + rs.pos_rel + 1,rs,pos_max,base_max,coln);
				break;
			case TEXT_KEY_IDLE_0 :
				menu_center(rand() % gc.size(),rs,pos_max,base_max,coln);
				done = true;
				break;
			case TEXT_KEY_IDLE_1 :
			case TEXT_KEY_HELP :
			case TEXT_KEY_GROUP :
			case TEXT_KEY_TYPE :
			case TEXT_KEY_EXCLUDE :
			case TEXT_KEY_SORT :
			case TEXT_KEY_SETGROUP :
			case TEXT_KEY_SETTYPE :
			case TEXT_KEY_ENTER :
			case TEXT_KEY_MODE :
			case TEXT_KEY_RUN_CLONE :
				done = true;
				break;
			case TEXT_KEY_ESC :
				++exit_count;
				if (exit_count >= rs.exit_count)
					done = true;
				break;
			default:
				if (isalnum(key)) {
					oldfast.insert(oldfast.length(),1,key);
					pgame_array::const_iterator i;
					for(i=gc.begin();i!=gc.end();++i) {
						if (menu_fast_compare((*i)->description_get(),oldfast))
							break;
					}
					if (i!=gc.end()) {
						int pos = i - gc.begin();
						menu_pos(pos,rs,pos_max,base_max,coln);
						rs.fast = oldfast;
					}
				}
		}

		if (rs.pos_rel + rs.pos_base < gc.size()) {
			rs.pgame = gc[rs.pos_rel+rs.pos_base];
			rs.pclone = 0;
		} else {
			rs.pgame = 0;
			rs.pclone = 0;
		}
	}

	if (rs.pgame)
		rs.last_game = rs.pgame->name_get();
	else
		rs.last_game = "";

	delete [] cell_map;

	return key;
}

int run_menu_idle(menu_state& rs, pgame_array& gc) {
	bool done = false;
	int key = 0;

	text_backdrop_init(COLOR_MENU_BACKDROP, 1);
	text_backdrop_set(0,0,0,text_dx_get(),text_dy_get());

	while (!done) {
		unsigned i = rand() % gc.size();

		text_clear();

		string backdrop;
		if (backdrop_find_preview(backdrop,rs,gc[i])) {
			text_backdrop_set(0, backdrop.c_str(), false );
		} else {
			text_backdrop_clear(0, false);
		}

		text_backdrop_update();

		key = text_getkey();

		if (key != TEXT_KEY_IDLE_1)
			done = true;

	}

	return key;
}

int run_menu(const game_set& gar, menu_state& rs, bool flipxy) {
	pgame_sort_set* psc;

	// sort
	switch (rs.sort_mode) {
		case sort_by_root_name :
			psc = new pgame_sort_set(sort_by_root_name_func);
			break;
		case sort_by_name :
			psc = new pgame_sort_set(sort_by_name_func);
			break;
		case sort_by_manufacturer :
			psc = new pgame_sort_set(sort_by_manufacturer_func);
			break;
		case sort_by_year :
			psc = new pgame_sort_set(sort_by_year_func);
			break;
		case sort_by_time :
			psc = new pgame_sort_set(sort_by_time_func);
			break;
		case sort_by_group :
			psc = new pgame_sort_set(sort_by_group_func);
			break;
		case sort_by_type :
			psc = new pgame_sort_set(sort_by_type_func);
			break;
		case sort_by_size :
			psc = new pgame_sort_set(sort_by_size_func);
			break;
		default:
			return TEXT_KEY_NONE;
	}


	int game_good = 0;
	for(game_set::const_iterator i=gar.begin();i!=gar.end();++i) {
		bool accept = true;
		bool accept_group = rs.include_group.size() == 0;
		bool accept_type = rs.include_type.size() == 0;

		// group
		for(group_container::iterator j=rs.include_group.begin();j!=rs.include_group.end();++j) {
			if (*j == i->group_derived_get()) {
				accept_group = true;
				break;
			}
		}

		// type
		for(type_container::iterator j=rs.include_type.begin();j!=rs.include_type.end();++j) {
			if (*j == i->type_derived_get()) {
				accept_type = true;
				break;
			}
		}

		if  (accept_group && accept_type) {
			bool present = false;
			bool good = false;

			// exclude
			if (accept && rs.exclude_clone == exclude && i->parent_get()!=0)
				accept = false;
			if (accept && rs.exclude_clone == exclude_not && i->parent_get()==0)
				accept = false;
			if (accept) {
				const game* gb;
			 	if (rs.exclude_clone == exclude)
 					gb = i->clone_best_get();
				else
					gb = &*i;
				good = gb->is_good();
			}
			if (accept && rs.exclude_bad == exclude && !good)
				accept = false;
			if (accept && rs.exclude_bad == exclude_not && good)
				accept = false;
			if (accept && rs.exclude_neogeo == exclude && gar.is_game_neogeo(i->name_get()))
				accept = false;
			if (accept && rs.exclude_neogeo == exclude_not && !gar.is_game_neogeo(i->name_get()))
				accept = false;
			if (accept)
				present = gar.is_tree_rom_of_present(i->name_get(),rs.merge);
			if (accept && rs.exclude_missing == exclude && !present)
				accept = false;
			if (accept && rs.exclude_missing == exclude_not && present)
				accept = false;
			if (accept && rs.exclude_vector == exclude && i->vector_get())
				accept = false;
			if (accept && rs.exclude_vector == exclude_not && !i->vector_get())
				accept = false;
			if (accept && i->resource_get())
				accept = false;
			if (accept) {
				if (present && good)
					++game_good;
				psc->insert(&*i);
			}
		}
	}

	pgame_array gc(psc->size());
	pgame_array::iterator j=gc.begin();
	for(pgame_sort_set::const_iterator i=psc->begin();i!=psc->end();++i,++j)
		*j = *i;
	delete psc;

	bool done = false;
	int key = 0;
	bool idle = false;

	while (!done) {
		if (idle)
			key = run_menu_idle(rs,gc);
		else
			key = run_menu_user(gar,rs,flipxy,gc,game_good);

		idle = key == TEXT_KEY_IDLE_1;

		switch (key) {
			case TEXT_KEY_MODE :
				switch (rs.mode) {
					case mode_list : rs.mode = mode_tile_3x2; break;
					case mode_tile_3x2 : rs.mode = mode_tile_4x3; break;
					case mode_tile_4x3 : rs.mode = mode_list; break;
				}
				break;
			case TEXT_KEY_IDLE_1 :
				break;
			default:
				done = true;
		}
	}

	return key;
}

// ------------------------------------------------------------------------
// Sort menu

#define SORT_CHOICE_X text_dx_get()/8
#define SORT_CHOICE_Y text_dy_get()/5
#define SORT_CHOICE_DX 20*text_font_dx_get()

void run_sort(menu_state& rs) {
	choice_container ch;
	choice_container::iterator pos[8];

	pos[0] = ch.insert( ch.end(), choice("Parent name") );
	pos[1] = ch.insert( ch.end(), choice("Name") );
	pos[2] = ch.insert( ch.end(), choice("Time played") );
	pos[3] = ch.insert( ch.end(), choice("Group") );
	pos[4] = ch.insert( ch.end(), choice("Type") );
	pos[5] = ch.insert( ch.end(), choice("Manufacturer") );
	pos[6] = ch.insert( ch.end(), choice("Year") );
	pos[7] = ch.insert( ch.end(), choice("Size") );

	choice_container::iterator i;
	switch (rs.sort_mode) {
		case sort_by_root_name: i = pos[0]; break;
		case sort_by_name: i = pos[1]; break;
		case sort_by_time: i = pos[2]; break;
		case sort_by_group: i = pos[3]; break;
		case sort_by_type: i = pos[4]; break;
		case sort_by_manufacturer: i = pos[5]; break;
		case sort_by_year: i = pos[6]; break;
		case sort_by_size: i = pos[7]; break;
	}

	int key = run_choice(" Sort mode",SORT_CHOICE_X,SORT_CHOICE_Y,SORT_CHOICE_DX,ch,i);
	if (key == TEXT_KEY_ENTER) {
		if (0) ;
		else if (i == pos[0]) rs.sort_mode = sort_by_root_name;
		else if (i == pos[1]) rs.sort_mode = sort_by_name;
		else if (i == pos[2]) rs.sort_mode = sort_by_time;
		else if (i == pos[3]) rs.sort_mode = sort_by_group;
		else if (i == pos[4]) rs.sort_mode = sort_by_type;
		else if (i == pos[5]) rs.sort_mode = sort_by_manufacturer;
		else if (i == pos[6]) rs.sort_mode = sort_by_year;
		else if (i == pos[7]) rs.sort_mode = sort_by_size;
	}
}

// ------------------------------------------------------------------------
// Exclude menu

#define EXCLUDE_CHOICE_X text_dx_get()/8
#define EXCLUDE_CHOICE_Y text_dy_get()/5
#define EXCLUDE_CHOICE_DX 30*text_font_dx_get()

void run_exclude(menu_state& rs) {
	choice_container ch;
	choice_container::iterator pos[16];

	pos[0] = ch.insert( ch.end(), choice("Preliminary", rs.exclude_bad) );
	pos[1] = ch.insert( ch.end(), choice("Missing", rs.exclude_missing) );
	pos[2] = ch.insert( ch.end(), choice("Clone", rs.exclude_clone) );
	pos[3] = ch.insert( ch.end(), choice("Neogeo", rs.exclude_neogeo) );
	pos[4] = ch.insert( ch.end(), choice("Vector", rs.exclude_vector) );

	choice_container::iterator i = ch.begin();;
	int key = run_choice(" Show games",EXCLUDE_CHOICE_X,EXCLUDE_CHOICE_Y,EXCLUDE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.exclude_bad = pos[0]->tristate;
		rs.exclude_missing = pos[1]->tristate;
		rs.exclude_clone = pos[2]->tristate;
		rs.exclude_neogeo = pos[3]->tristate;
		rs.exclude_vector = pos[4]->tristate;
	}
}

// ------------------------------------------------------------------------
// Group menu

#define GROUP_CHOICE_X text_dx_get()/8
#define GROUP_CHOICE_Y text_dy_get()/5
#define GROUP_CHOICE_DX 20*text_font_dx_get()

void run_group(menu_state& rs, const group_container& grp) {
	choice_container ch;
	choice_container::iterator pos[32];

	int jn = 0;
	for(group_container::const_iterator j = grp.begin();j!=grp.end() && jn<32;++j) {
		bool tag = false;
		for(group_container::const_iterator k = rs.include_group.begin();k!=rs.include_group.end();++k) {
			if (*j == *k) {
				tag = true;
				break;
			}
		}
		pos[jn] = ch.insert( ch.end(), choice(*j, tag) );
		++jn;
	}

	choice_container::iterator i = ch.begin();;
	int key = run_choice(" Show groups",GROUP_CHOICE_X,GROUP_CHOICE_Y,GROUP_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.include_group.erase(rs.include_group.begin(), rs.include_group.end());
		for(choice_container::const_iterator j=ch.begin();j!=ch.end();++j) {
			if (j->bistate)
				rs.include_group.insert( rs.include_group.end(), j->desc );
		}
	}
}

// ------------------------------------------------------------------------
// Type menu

#define TYPE_CHOICE_X text_dx_get()/8
#define TYPE_CHOICE_Y 2*text_font_dy_get()
#define TYPE_CHOICE_DX 30*text_font_dx_get()

void run_type(menu_state& rs, const type_container& typ) {
	choice_container ch;
	choice_container::iterator pos[32];

	int jn = 0;
	for(type_container::const_iterator j = typ.begin();j!=typ.end() && jn<32;++j) {
		bool tag = false;
		for(type_container::const_iterator k = rs.include_type.begin();k!=rs.include_type.end();++k) {
			if (*j == *k) {
				tag = true;
				break;
			}
		}
		pos[jn] = ch.insert( ch.end(), choice(*j, tag) );
		++jn;
	}

	choice_container::iterator i = ch.begin();;
	int key = run_choice(" Show types",TYPE_CHOICE_X,TYPE_CHOICE_Y,TYPE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.include_type.erase(rs.include_type.begin(), rs.include_type.end());
		for(choice_container::const_iterator j=ch.begin();j!=ch.end();++j) {
			if (j->bistate)
				rs.include_type.insert( rs.include_type.end(), j->desc );
		}
	}
}

// ------------------------------------------------------------------------
// Move menu

void run_group_move(menu_state& rs, const group_container& grp) {
	choice_container ch;
	choice_container::iterator i = ch.begin();

	for(group_container::const_iterator j = grp.begin();j!=grp.end();++j) {
		choice_container::iterator k = ch.insert( ch.end(), choice(*j) );
		if (rs.pgame->group_get() == *j)
			i = k;
	}

	int key = run_choice(" Select group",GROUP_CHOICE_X,GROUP_CHOICE_Y,GROUP_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.pgame->group_set( i->desc );
	}
}

void run_type_move(menu_state& rs, const type_container& typ) {
	choice_container ch;
	choice_container::iterator i = ch.begin();

	for(type_container::const_iterator j = typ.begin();j!=typ.end();++j) {
		choice_container::iterator k = ch.insert( ch.end(), choice(*j) );
		if (rs.pgame->type_get() == *j)
			i = k;
	}

	int key = run_choice(" Select type",TYPE_CHOICE_X,TYPE_CHOICE_Y,TYPE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		rs.pgame->type_set( i->desc );
	}
}

// ------------------------------------------------------------------------
// Clone menu

#define CLONE_CHOICE_X text_dx_get()/10
#define CLONE_CHOICE_Y text_dy_get()/5
#define CLONE_CHOICE_DX text_dx_get()*4/5

void run_clone(menu_state& rs) {
	choice_container ch;
	choice_container::iterator pos[32];
	const game* game[32];
	int jn = 0;
	choice_container::iterator i;

	rs.pclone = 0;
	ostrstream s;
	s << rs.pgame->description_get() << ", " << rs.pgame->manufacturer_get() << ", " << rs.pgame->year_get() << ends;
	pos[jn] = ch.insert( ch.end(), choice( s.str() ) );
	s.freeze(0);
	game[jn] = rs.pgame;
	++jn;

	for(pgame_container::const_iterator j = rs.pgame->clone_bag_get().begin();j!=rs.pgame->clone_bag_get().end();++j) {
		if (jn < 32) {
			ostrstream s;
			s << (*j)->description_get() << ", " << (*j)->manufacturer_get() << ", " << (*j)->year_get() << ends;
			pos[jn] = ch.insert( ch.end(), choice( s.str() ) );
			s.freeze(0);
			game[jn] = &**j;
			if ((*j)->name_get() == rs.pgame->name_get())
				i = pos[jn];
			++jn;
		}
	}

	i = pos[0];

	int key = run_choice(" Select clone",CLONE_CHOICE_X,CLONE_CHOICE_Y,CLONE_CHOICE_DX,ch,i);

	if (key == TEXT_KEY_ENTER) {
		int j;
		for(j=0;j<jn;++j) {
			if (pos[j] == i) {
				rs.pclone = game[j];
				return;
			}
		}
	}

}

// ------------------------------------------------------------------------
// Help menu

void run_help() {
	text_backdrop_clear();

	text_clear(0,0,text_dx_get(),text_dy_get(),COLOR_HELP_NORMAL >> 4);
	text_clear(0,0,text_dx_get(),text_font_dy_get(),COLOR_MENU_BAR >> 4);
	text_put(2*text_font_dx_get(),0,"HELP",COLOR_MENU_BAR);

	int y = 2*text_font_dy_get();
	int xt = 1*text_font_dx_get();
	int xd = 8*text_font_dx_get();
	text_put(xt,y,"F1",COLOR_HELP_TAG);
	text_put(xd,y,"Help screen",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"F2",COLOR_HELP_TAG);
	text_put(xd,y,"Include/Exclude specific games by group",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"F3",COLOR_HELP_TAG);
	text_put(xd,y,"Include/Exclude specific games by type",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"F4",COLOR_HELP_TAG);
	text_put(xd,y,"Include/Exclude specific games by attribute",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"F5",COLOR_HELP_TAG);
	text_put(xd,y,"Select the game sort method",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"F9",COLOR_HELP_TAG);
	text_put(xd,y,"Change the current game group",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"F10",COLOR_HELP_TAG);
	text_put(xd,y,"Change the current game type",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"F12",COLOR_HELP_TAG);
	text_put(xd,y,"Run a clone",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"ENTER",COLOR_HELP_TAG);
	text_put(xd,y,"Run the current game",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"TAB",COLOR_HELP_TAG);
	text_put(xd,y,"Change the menu mode",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"SPACE",COLOR_HELP_TAG);
	text_put(xd,y,"Change the preview type snap/flyers",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"ESC",COLOR_HELP_TAG);
	text_put(xd,y,"Exit",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	y += text_font_dy_get();
	text_put(xt,y,"In the selection submenus:",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"ENTER",COLOR_HELP_TAG);
	text_put(xd,y,"Accept",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"-",COLOR_HELP_TAG);
	text_put(xd,y,"Unselect all",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"+",COLOR_HELP_TAG);
	text_put(xd,y,"Select all",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"SPACE",COLOR_HELP_TAG);
	text_put(xd,y,"Toggle (+ include, - exclude, * required)",COLOR_HELP_NORMAL);
	y += text_font_dy_get();
	text_put(xt,y,"ESC",COLOR_HELP_TAG);
	text_put(xd,y,"Cancel",COLOR_HELP_NORMAL);

	text_getkey();
}

