/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "text.h"
#include "font.h"
#include "backdrop.h"
#include "screen.h"
#include "vga.h"
#include "vbe.h"
#include "allegro.h"

#include <iostream.h>
#include <unistd.h>
#include <math.h>
#include <sys/farptr.h>
#include <sys/segments.h>
#include <sys/stat.h>
#include <conio.h>
#include <stdio.h>
#include <pc.h>

#define VIDEO_MAX_X 1600
#define VIDEO_MAX_Y 1280

// Orientation flag
static unsigned text_orientation = 0;

// Font (already orientation corrected)
static unsigned real_font_dx;
static unsigned real_font_dy;
static struct bitmapfont* real_font_map;

struct backdrop_data {
	// Backdrop (already orientation corrected)
	struct bitmap* backdrop;
	char backdrop_path[256];

	// RGB palette for 8 bit backdrop
	video_rgb backdrop_color_map[256];
	unsigned backdrop_index_map[256];

	// RGB conversion table
	unsigned backdrop_convert[BITMAP_INDEX_MAX];

	// Backdrop expansion factor (also used for MIRROR effect)
	unsigned backdrop_x_map[VIDEO_MAX_X];
	unsigned backdrop_y_map[VIDEO_MAX_Y];

	// Position of the backdrop in the screen
	int backdrop_x;
	int backdrop_y;
	int backdrop_dx;
	int backdrop_dy;

	// Position of the backdrop in the screen, already rotated
	int real_backdrop_x;
	int real_backdrop_y;
	int real_backdrop_dx;
	int real_backdrop_dy;

	bool highlight;
};

// Color used for missing backdrop
static unsigned backdrop_missing_color;

#define BACKDROP_MAX 12

static unsigned backdrop_mac;

static struct backdrop_data BACKDROP[BACKDROP_MAX];

// -----------------------------------------------------------------------
// Joystick

struct { char *name; int id; } joy_table[] =
{
	{ "none",                       JOY_TYPE_NONE },
	{ "auto",                       JOY_TYPE_AUTODETECT },
	{ "standard",           JOY_TYPE_STANDARD },
	{ "dual",                       JOY_TYPE_2PADS },
	{ "4button",            JOY_TYPE_4BUTTON },
	{ "6button",            JOY_TYPE_6BUTTON },
	{ "8button",            JOY_TYPE_8BUTTON },
	{ "fspro",                      JOY_TYPE_FSPRO },
	{ "wingex",                     JOY_TYPE_WINGEX },
	{ "sidewinder",         JOY_TYPE_SIDEWINDER },
	{ "gamepadpro",         JOY_TYPE_GAMEPAD_PRO },
	{ "grip",               JOY_TYPE_GRIP },
	{ "grip4",              JOY_TYPE_GRIP4 },
	{ "sneslpt1",           JOY_TYPE_SNESPAD_LPT1 },
	{ "sneslpt2",           JOY_TYPE_SNESPAD_LPT2 },
	{ "sneslpt3",           JOY_TYPE_SNESPAD_LPT3 },
	{ "psxlpt1",            JOY_TYPE_PSXPAD_LPT1 },
	{ "psxlpt2",            JOY_TYPE_PSXPAD_LPT2 },
	{ "psxlpt3",            JOY_TYPE_PSXPAD_LPT3 },
	{ "n64lpt1",            JOY_TYPE_N64PAD_LPT1 },
	{ "n64lpt2",            JOY_TYPE_N64PAD_LPT2 },
	{ "n64lpt3",            JOY_TYPE_N64PAD_LPT3 },
	{ "wingwarrior",        JOY_TYPE_WINGWARRIOR },
	{ 0, 0 }
};

static int joystick = -1;

static void joystick_init() {
	char* joyname = get_config_string("config", "joystick", 0);
	if (!joyname)
		joyname = get_config_string("config", "joy", "none");

	unsigned i;
	for(i=0;joy_table[i].name;++i)
		if (strcmp(joy_table[i].name,joyname)==0)
			break;

	if (joy_table[i].name) {
		joystick = joy_table[i].id;
		if (install_joystick(joystick)!=0) {
			joystick = -1;
		}
	}
}

static void joystick_done() {
	if (joystick>=0)
		remove_joystick();
}

static int joystick_raw_poll() {
	if (joystick>=0) {
		poll_joystick();

		for(int i=0;i<num_joysticks;++i) {
			for(int j=0;j<joy[i].num_sticks;++j) {
				if (joy[i].stick[j].num_axis > 0) {
					if (joy[i].stick[j].axis[0].d1)
						return TEXT_KEY_LEFT;
					if (joy[i].stick[j].axis[0].d2)
						return TEXT_KEY_RIGHT;
				}
				if (joy[i].stick[j].num_axis > 1) {
					if (joy[i].stick[j].axis[1].d1)
						return TEXT_KEY_UP;
					if (joy[i].stick[j].axis[1].d2)
						return TEXT_KEY_DOWN;
				}
			}
			for(int j=0;j<joy[i].num_buttons;++j) {
				if (joy[i].button[j].b) {
					switch (j) {
						case 0 : return TEXT_KEY_ENTER;
						case 1 : return TEXT_KEY_ESC;
						case 2 : return TEXT_KEY_SPACE;
						case 3 : return TEXT_KEY_MODE;
					}
				}
			}
		}

	}
	return TEXT_KEY_NONE;
}

// -------------------------------------------------------------------------
// Size

__inline__ void swap(unsigned& a, unsigned& b) {
	unsigned t = a;
	a = b;
	b = t;
}

__inline__ void swap(int& a, int& b) {
	int t = a;
	a = b;
	b = t;
}

int text_dx_get() {
	if (text_orientation & ORIENTATION_FLIP_XY)
		return video_size_y();
	else
		return video_size_x();
}

int text_dy_get() {
	if (text_orientation & ORIENTATION_FLIP_XY)
		return video_size_x();
	else
		return video_size_y();
}

int text_font_dx_get() {
	if (text_orientation & ORIENTATION_FLIP_XY)
		return real_font_dy;
	else
		return real_font_dx;
}

int text_font_dy_get() {
	if (text_orientation & ORIENTATION_FLIP_XY)
		return real_font_dx;
	else
		return real_font_dy;
}

// -------------------------------------------------------------------------
// Video mode choice

#define DEFAULT_GRAPH_MODE "default_graph" // default video mode

// comparing for graphics mode
bool mode_graphics_less(const video_mode* A, const video_mode* B) {
	int bitsA = video_mode_bytes_per_pixel(A);
	int bitsB = video_mode_bytes_per_pixel(B);
	if (bitsA < bitsB) return true;
	if (bitsA > bitsB) return false;

	int areaA = A->size_x * A->size_y;
	int areaB = B->size_x * B->size_y;
	int difA = abs( areaA - 640*480 );
	int difB = abs( areaB - 640*480 );
	if (difA > difB) return true;
	if (difA < difB) return false;

	return false;
}

static video_mode mode; // selected video mode
static video_monitor monitor; // monitor info

// -------------------------------------------------------------------------
// Inizialization

bool text_init() {
	char* cfg = "mame.cfg";

	bool mode_found = false; // selected video mode found
	video_mode_container vm_sel;
	video_mode_container vm_def;
	video_mode_container_iterator i;

	if (allegro_init() != 0) {
		cerr << "Error inizializing the Allegro library" << endl;
		goto out;
	}
	set_config_file(cfg);

	joystick_init();

	if (video_options_load() != 0) {
		cerr << "Error loading the video options" << endl << video_error_description_get() << endl;
		goto out_joy;
	}

	video_init();

	if (video_monitor_load(&monitor) < 0) {	// < 0, ignore missing hclock and vclock specifications
		cerr << "Error loading the monitor options" << endl << video_error_description_get() << endl;
		goto out_video;
	}

	// load selected
	video_mode_container_init(&vm_sel);
	video_mode_container_load(&vm_sel);

	// load default
	video_mode_container_init(&vm_def);
	video_mode_container_insert_vga(&vm_def);
	video_mode_container_insert_vbe(&vm_def);

	// seach default mode in selected
	for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
		const video_mode* vm = video_mode_container_iterator_get(&i);
		if (video_mode_is_graphics(vm)
			&& strcmp(vm->name,DEFAULT_GRAPH_MODE)==0) {
			mode = *vm;
			mode_found = true;
			break;
		}
	}

	// search graphics mode in selected
	if (!mode_found) {
		for(video_mode_container_iterator_begin(&i,&vm_sel);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
			const video_mode* vm = video_mode_container_iterator_get(&i);
			if (video_mode_is_graphics(vm)
				&& video_monitor_mode_check(&monitor,vm)
				&& (video_mode_bytes_per_pixel(vm)==1 || video_mode_bytes_per_pixel(vm)==2)) {
				if (!mode_found || mode_graphics_less(&mode,vm)) {
					mode = *vm;
					mode_found = true;
				}
			}
		}
	}

	// search graphics mode in default
	if (!mode_found) {
		for(video_mode_container_iterator_begin(&i,&vm_def);!video_mode_container_iterator_end(&i);video_mode_container_iterator_next(&i)) {
			const video_mode* vm = video_mode_container_iterator_get(&i);
			if (video_mode_is_graphics(vm)
				&& video_monitor_mode_check(&monitor,vm)
				&& (video_mode_bytes_per_pixel(vm)==1 || video_mode_bytes_per_pixel(vm)==2)) {
				if (!mode_found || mode_graphics_less(&mode,vm)) {
					mode = *vm;
					mode_found = true;
				}
			}
		}
	}

	video_mode_container_done(&vm_def);
	video_mode_container_done(&vm_sel);

	if (!mode_found) {
		cerr << "No video mode avaliable" << endl;
		goto out_video;
	}

	return true;

out_video:
	video_done();
out_joy:
	joystick_done();
	allegro_exit();
out:
	return false;
}

void text_done(void) {
	video_mode_reset();
	video_done();
	allegro_exit();
}

static double text_gamma;
static double text_brightness;
static unsigned text_idle_0;
static unsigned text_idle_0_rep;
static unsigned text_idle_1;
static unsigned text_idle_1_rep;
static clock_t text_idle_clock;

static unsigned video_buffer_size;
static unsigned video_buffer_line_size;
static unsigned video_buffer_pixel_size;
static unsigned char* video_buffer;

static void palette_set(video_color* pal, unsigned pal_max, unsigned* conv);

// Inizialize the submenu
bool text_sub_init(const char* font, unsigned orientation, double gamma, double brightness, unsigned idle_0, unsigned idle_0_rep,unsigned idle_1, unsigned idle_1_rep) {
	text_idle_clock = clock();
	text_idle_0 = idle_0;
	text_idle_1 = idle_1;
	text_idle_0_rep = idle_0_rep;
	text_idle_1_rep = idle_1_rep;

	text_orientation = orientation;
	if (gamma < 0.1) gamma = 0.1;
	if (gamma > 10) gamma = 10;
	text_gamma = 1.0 / gamma;
	text_brightness = brightness;

	// load the font
	real_font_map = bitmapfont_load(font);
	if (!real_font_map)
		real_font_map = bitmapfont_default();

	// set the orientation
	bitmapfont_orientation(real_font_map,text_orientation,0);

	// compute font size
	real_font_dx = bitmapfont_size_x(real_font_map);
	real_font_dy = bitmapfont_size_y(real_font_map);

	// startup keyboard
	if (install_keyboard() != 0) {
		goto out_font;
	}

	// set the video mode
	if (video_mode_set(&mode) != 0) {
		goto out_key;
	}

	video_buffer_pixel_size = video_bytes_per_pixel();
	video_buffer_line_size = video_size_x() * video_bytes_per_pixel();
	video_buffer_size =  video_size_y() * video_buffer_line_size;
	video_buffer = (unsigned char*)operator new(video_buffer_size);
	memset(video_buffer,0,video_buffer_size);


	// clear the screen
	for(unsigned y=0;y<video_size_y();++y)
		for(unsigned x=0;x<video_size_x();++x)
			video_put_pixel(x,y,0);

	// set font
	bitmapfont_set(real_font_map);

	if (video_index() == VIDEO_FLAGS_INDEX_PACKED)
		palette_set(0,0,0);

	return true;

out_key:
	remove_keyboard();
out_font:
	bitmapfont_free(real_font_map);
	return false;
}

// Deinizialize the sub menu
void text_sub_done(void) {
	video_mode_reset();
	remove_keyboard();
	bitmapfont_free(real_font_map);
	operator delete(video_buffer);
}

//---------------------------------------------------------------------------
// Color
#define PALETTE_RESERVED_MAX 16

static video_color PALETTE_RESERVED[PALETTE_RESERVED_MAX] = {
{   0,  0,  0,0 },
{ 192,  0,  0,0 },
{   0,192,  0,0 },
{ 192,192,  0,0 },
{   0,  0,192,0 },
{ 192,  0,192,0 },
{   0,192,192,0 },
{ 192,192,192,0 },
{ 128,128,128,0 },
{ 255,  0,  0,0 },
{   0,255,  0,0 },
{ 255,255,  0,0 },
{   0,  0,255,0 },
{ 255,  0,255,0 },
{   0,255,255,0 },
{ 255,255,255,0 }
};

static video_color text_palette[256];

int palette_dist(const video_color& A, const video_color& B) {
	int r = (int)A.red - B.red;
	int g = (int)A.green - B.green;
	int b = (int)A.blue - B.blue;
	return r*r + g*g + b*b;
}

static void palette_set(video_color* pal, unsigned pal_max, unsigned* conv) {
	unsigned i;
	for(i=0;i<PALETTE_RESERVED_MAX;++i)
		text_palette[i] = PALETTE_RESERVED[i];

	for(i=0;i<256-PALETTE_RESERVED_MAX && i<pal_max;++i) {
		text_palette[i+PALETTE_RESERVED_MAX] = pal[i];
		conv[i] = i+PALETTE_RESERVED_MAX;
	}

	// search for best colors
	while (i<pal_max) {
		unsigned best = PALETTE_RESERVED_MAX;
		int dist = palette_dist(pal[i],text_palette[best]);
		for(unsigned j=best+1;j<256;++j) {
			int ndist = palette_dist(pal[i],text_palette[j]);
			if (ndist < dist) {
				dist = ndist;
				best = j;
			}
		}
		conv[i] = best;
		++i;
	}

	for(i=0;i<256;++i) {
		double f = 255.0 * text_brightness;
		text_palette[i].red = (uint8)(f * pow(text_palette[i].red / 255.0, text_gamma));
		text_palette[i].green = (uint8)(f * pow(text_palette[i].green / 255.0, text_gamma));
		text_palette[i].blue = (uint8)(f * pow(text_palette[i].blue / 255.0, text_gamma));
	}
}

static inline unsigned color2palette(unsigned color) {
	return color & 0xF;
}

// Conversion from text color to video color
static inline video_rgb color2rgb(unsigned color) {
	video_rgb r;
	color &= 0xF;
	video_rgb_make(&r, PALETTE_RESERVED[color].red, PALETTE_RESERVED[color].green, PALETTE_RESERVED[color].blue);
	return r;
}


//---------------------------------------------------------------------------
// Put

static void text_put16rgb_char_font(unsigned x, unsigned y, unsigned bitmap, unsigned attrib) {
	struct bitmap* src = real_font_map->data[bitmap];
	video_rgb color_fore = color2rgb(attrib);
	video_rgb color_back = color2rgb(attrib >> 4);
	unsigned char* buffer = video_buffer + x * video_buffer_pixel_size + y * video_buffer_line_size;
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned char* src_ptr = bitmap_line(src,cy);
		unsigned short* dst_ptr = (unsigned short*)buffer;
		for(unsigned cx=0;cx<real_font_dx;++cx) {
			unsigned color = *src_ptr ? color_fore : color_back;
			*dst_ptr = color;
			dst_ptr += 1;
			src_ptr += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_put8rgb_char_font(unsigned x, unsigned y, unsigned bitmap, unsigned attrib) {
	struct bitmap* src = real_font_map->data[bitmap];
	unsigned color_fore = color2rgb(attrib);
	unsigned color_back = color2rgb(attrib >> 4);
	unsigned char* buffer = video_buffer + x * video_buffer_pixel_size + y * video_buffer_line_size;
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned char* src_ptr = bitmap_line(src,cy);
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<real_font_dx;++cx) {
			unsigned color = *src_ptr ? color_fore : color_back;
			*dst_ptr = color;
			dst_ptr += 1;
			src_ptr += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_put8packed_char_font(unsigned x, unsigned y, unsigned bitmap, unsigned attrib) {
	struct bitmap* src = real_font_map->data[bitmap];
	unsigned color_fore = color2palette(attrib);
	unsigned color_back = color2palette(attrib >> 4);
	unsigned char* buffer = video_buffer + x * video_buffer_pixel_size + y * video_buffer_line_size;
	for(unsigned cy=0;cy<real_font_dy;++cy) {
		unsigned char* src_ptr = bitmap_line(src,cy);
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<real_font_dx;++cx) {
			unsigned color = *src_ptr ? color_fore : color_back;
			*dst_ptr = color;
			dst_ptr += 1;
			src_ptr += 1;
		}
		buffer += video_buffer_line_size;
	}
}

void text_put(int x, int y, char c, int color) {
	if (x>=0 && y>=0 && x+text_font_dx_get()<=text_dx_get() && y+text_font_dy_get()<=text_dy_get()) {
		if (text_orientation & ORIENTATION_FLIP_XY)
			swap(x,y);
		if (text_orientation & ORIENTATION_MIRROR_X)
			x = video_size_x() - real_font_dx - x;
		if (text_orientation & ORIENTATION_MIRROR_Y)
			y = video_size_y() - real_font_dy - y;

		assert( x>=0 && y>=0 && x+real_font_dx<=video_size_x() &&  y+real_font_dy<=video_size_y() );

		if (video_bytes_per_pixel() == 1) {
			if (video_index() == VIDEO_FLAGS_INDEX_PACKED)
				text_put8packed_char_font(x,y,(unsigned char)c,color);
			else
				text_put8rgb_char_font(x,y,(unsigned char)c,color);
		} else {
			text_put16rgb_char_font(x,y,(unsigned char)c,color);
		}
	}
}

static void text_clear16rgb(unsigned x, unsigned y, unsigned dx, unsigned dy, unsigned attrib) {
	video_rgb color = color2rgb(attrib);
	unsigned char* buffer = video_buffer + x * video_buffer_pixel_size + y * video_buffer_line_size;
	for(unsigned cy=0;cy<dy;++cy) {
		unsigned short* dst_ptr = (unsigned short*)buffer;
		for(unsigned cx=0;cx<dx;++cx) {
			*dst_ptr = color;
			dst_ptr += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_clear8rgb(unsigned x, unsigned y, unsigned dx, unsigned dy, unsigned attrib) {
	unsigned color = color2rgb(attrib);
	unsigned char* buffer = video_buffer + x * video_buffer_pixel_size + y * video_buffer_line_size;
	for(unsigned cy=0;cy<dy;++cy) {
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<dx;++cx) {
			*dst_ptr = color;
			dst_ptr += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_clear8packed(unsigned x, unsigned y, unsigned dx, unsigned dy, unsigned attrib) {
	unsigned color = color2palette(attrib);
	unsigned char* buffer = video_buffer + x * video_buffer_pixel_size + y * video_buffer_line_size;
	for(unsigned cy=0;cy<dy;++cy) {
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<dx;++cx) {
			*dst_ptr = color;
			dst_ptr += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_clear_noclip(int x, int y, int dx, int dy, int color) {
	if (text_orientation & ORIENTATION_FLIP_XY) {
		swap(x,y);
		swap(dx,dy);
	}
	if (text_orientation & ORIENTATION_MIRROR_X)
		x = video_size_x() - dx - x;
	if (text_orientation & ORIENTATION_MIRROR_Y)
		y = video_size_y() - dy - y;

	assert( x>=0 && y>=0 && x+dx<=video_size_x() &&  y+dy<=video_size_y() );

	if (video_bytes_per_pixel() == 1) {
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED)
			text_clear8packed(x,y,dx,dy,color);
		else
			text_clear8rgb(x,y,dx,dy,color);
	} else {
		text_clear16rgb(x,y,dx,dy,color);
	}
}

void text_clear(int x, int y, int dx, int dy, int color) {
	if (x < 0) {
		dx += x;
		x = 0;
	}
	if (y < 0) {
		dy += y;
		y = 0;
	}
	if (x + dx > text_dx_get()) {
		dx = text_dx_get() - x;
	}
	if (y + dy > text_dy_get()) {
		dy = text_dy_get() - y;
	}
	if (dx > 0 && dy > 0)
		text_clear_noclip(x,y,dx,dy,color);
}

void text_box(int x, int y, int dx, int dy, int color) {
	text_clear(x,y,dx,1,color);
	text_clear(x,y+dy-1,dx,1,color);
	text_clear(x,y+1,1,dy-2,color);
	text_clear(x+dx-1,y+1,1,dy-2,color);
}

void text_put(int x, int y, const string& s, int color) {
	for(unsigned i=0;i<s.length();++i)
		text_put(x+i*text_font_dx_get(),y,s[i],color);
}

void text_fill(int x, int y, int dx, char c, int color) {
	for(int i=0;i<dx;++i)
		text_put(x+i*text_font_dx_get(),y,c,color);
}

//---------------------------------------------------------------------------
// Backdrop

// Select the backdrop
void text_backdrop_set(int back_index, const char* file, bool highlight) {
	if (strcmp(BACKDROP[back_index].backdrop_path,file) != 0
		|| BACKDROP[back_index].highlight != highlight) {
		strcpy(BACKDROP[back_index].backdrop_path,file);
		BACKDROP[back_index].highlight = highlight;
		if (BACKDROP[back_index].backdrop) {
			bitmap_free(BACKDROP[back_index].backdrop);
			BACKDROP[back_index].backdrop = 0;
		}
	}
}

// Clear the backdrop
void text_backdrop_clear(int back_index, bool highlight) {
	BACKDROP[back_index].backdrop_path[0] = 0;
	if (BACKDROP[back_index].backdrop) {
		bitmap_free(BACKDROP[back_index].backdrop);
		BACKDROP[back_index].backdrop = 0;
	}
	BACKDROP[back_index].highlight = highlight;
}

void text_backdrop_clear() {
	for(int i=0;i<BACKDROP_MAX;++i)
		text_backdrop_clear(i,false);
	backdrop_mac = 0;
}

void text_backdrop_init(unsigned Abackdrop_missing_color, unsigned Amac) {
	for(int i=0;i<BACKDROP_MAX;++i)
		text_backdrop_clear(i,false);

	backdrop_missing_color = Abackdrop_missing_color;
	backdrop_mac = Amac;

	if (backdrop_mac == 1 && video_index_rgb_to_packed_avaliable())
		video_index_rgb_to_packed();
	if (backdrop_mac > 1 && video_index_packed_to_rgb_avaliable())
		video_index_packed_to_rgb(0);
}

// Set the backdrop position and size
void text_backdrop_set(int back_index, int x, int y, int dx, int dy) {
	if (back_index >= backdrop_mac)
		backdrop_mac = back_index + 1;

	// shrink by 1 for the outline box
	BACKDROP[back_index].backdrop_x = BACKDROP[back_index].real_backdrop_x = x+1;
	BACKDROP[back_index].backdrop_y = BACKDROP[back_index].real_backdrop_y = y+1;
	BACKDROP[back_index].backdrop_dx = BACKDROP[back_index].real_backdrop_dx = dx-2;
	BACKDROP[back_index].backdrop_dy = BACKDROP[back_index].real_backdrop_dy = dy-2;

	if (text_orientation & ORIENTATION_FLIP_XY) {
		swap(BACKDROP[back_index].real_backdrop_x,BACKDROP[back_index].real_backdrop_y);
		swap(BACKDROP[back_index].real_backdrop_dx,BACKDROP[back_index].real_backdrop_dy);
	}
	if (text_orientation & ORIENTATION_MIRROR_X) {
		BACKDROP[back_index].real_backdrop_x = video_size_x() - BACKDROP[back_index].real_backdrop_x - BACKDROP[back_index].real_backdrop_dx;
	}
	if (text_orientation & ORIENTATION_MIRROR_Y) {
		BACKDROP[back_index].real_backdrop_y = video_size_y() - BACKDROP[back_index].real_backdrop_y - BACKDROP[back_index].real_backdrop_dy;
	}
}

//---------------------------------------------------------------------------
// Update

static void text_put8rgb_backdrop8(struct backdrop_data* back) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned char* back_ptr = bitmap_line(back->backdrop,back->backdrop_y_map[cy]);
		unsigned* bx = back->backdrop_x_map;
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
			unsigned color = back->backdrop_color_map[back_ptr[*bx]];
			*dst_ptr = color;
			dst_ptr += 1;
			bx += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_put8pack_backdrop8(struct backdrop_data* back) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned char* back_ptr = bitmap_line(back->backdrop,back->backdrop_y_map[cy]);
		unsigned* bx = back->backdrop_x_map;
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
			unsigned color = back->backdrop_index_map[back_ptr[*bx]];
			*dst_ptr = color;
			dst_ptr += 1;
			bx += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_put8rgb_backdrop24(struct backdrop_data* back) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned char* back_ptr = bitmap_line(back->backdrop,back->backdrop_y_map[cy]);
		unsigned* bx = back->backdrop_x_map;
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
			unsigned off = *bx * 3;
			video_rgb color;
			video_rgb_make(&color,back_ptr[off],back_ptr[off+1],back_ptr[off+2]);
			*dst_ptr = color;
			dst_ptr += 1;
			bx += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_put8pack_backdrop24(struct backdrop_data* back) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned char* back_ptr = bitmap_line(back->backdrop,back->backdrop_y_map[cy]);
		unsigned* bx = back->backdrop_x_map;
		unsigned char* dst_ptr = buffer;
		for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
			unsigned off = *bx * 3;
			unsigned color = back->backdrop_convert[BITMAP_COLOR_TO_INDEX(back_ptr[off],back_ptr[off+1],back_ptr[off+2])];
			*dst_ptr = color;
			dst_ptr += 1;
			bx += 1;
		}
		buffer += video_buffer_line_size;
	}
}

static void text_put16rgb_backdrop8(struct backdrop_data* back) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned char* back_ptr = bitmap_line(back->backdrop,back->backdrop_y_map[cy]);
		unsigned* bx = back->backdrop_x_map;
		unsigned short* dst_ptr = (unsigned short*)buffer;
		for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
			unsigned color = back->backdrop_color_map[back_ptr[*bx]];
			*dst_ptr = color;
			dst_ptr += 1;
			bx += 1;
		}
		buffer += video_buffer_line_size;
	}
}

void text_put16rgb_backdrop24(struct backdrop_data* back) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned char* back_ptr = bitmap_line(back->backdrop,back->backdrop_y_map[cy]);
		unsigned* bx = back->backdrop_x_map;
		unsigned short* dst_ptr = (unsigned short*)buffer;
		for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
			unsigned off = *bx * 3;
			video_rgb color;
			video_rgb_make(&color,back_ptr[off],back_ptr[off+1],back_ptr[off+2]);
			*dst_ptr = color;
			dst_ptr += 1;
			bx += 1;
		}
		buffer += video_buffer_line_size;
	}
}

/*
#define BOX_SIZE (video_size_x()/200+1)

static void text_put8_box(struct backdrop_data* back, unsigned color) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned char* dst_ptr = buffer;
		if (cy < BOX_SIZE || cy + BOX_SIZE > back->real_backdrop_dy) {
			for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
				*dst_ptr = color;
				dst_ptr += 1;
			}
		} else {
			for(unsigned cx=0;cx < BOX_SIZE;++cx) {
				*dst_ptr = color;
				dst_ptr += 1;
			}
			dst_ptr += back->real_backdrop_dx - BOX_SIZE*2;
			for(unsigned cx=0;cx < BOX_SIZE;++cx) {
				*dst_ptr = color;
				dst_ptr += 1;
			}
		}
		buffer += video_buffer_line_size;
	}
}

static void text_put16_box(struct backdrop_data* back, unsigned color) {
	unsigned char* buffer = video_buffer + back->real_backdrop_y * video_buffer_line_size + back->real_backdrop_x * video_buffer_pixel_size;
	for(unsigned cy=0;cy<back->real_backdrop_dy;++cy) {
		unsigned short* dst_ptr = (unsigned short*)buffer;
		if (cy < BOX_SIZE || cy + BOX_SIZE >= back->real_backdrop_dy) {
			for(unsigned cx=0;cx<back->real_backdrop_dx;++cx) {
				*dst_ptr = color;
				dst_ptr += 1;
			}
		} else {
			for(unsigned cx=0;cx < BOX_SIZE;++cx) {
				*dst_ptr = color;
				dst_ptr += 1;
			}
			dst_ptr += back->real_backdrop_dx - BOX_SIZE*2;
			for(unsigned cx=0;cx < BOX_SIZE;++cx) {
				*dst_ptr = color;
				dst_ptr += 1;
			}
		}
		buffer += video_buffer_line_size;
	}
}
*/

static void text_backdrop_cut(struct bitmap* bitmap, unsigned& _cx, unsigned& _cy) {
	unsigned yu = 0;
	unsigned yd = bitmap->size_y - 1;

	unsigned cx = bitmap->size_x / 2;
	unsigned cy = bitmap->size_y / 2;

	if (bitmap->bytes_per_pixel == 1) {
		uint8 c = *(uint8*)bitmap_line(bitmap,0);
		while (yu < yd && cx) {
			unsigned i;
			uint8* pu = (uint8*)bitmap_line(bitmap,yu);
			uint8* pd = (uint8*)bitmap_line(bitmap,yd);

			i = 0;
			while (i < cx && pu[i]==c && pu[bitmap->size_x - i -1]==c)
				++i;
			cx = i;

			i = 0;
			while (i < cx && pd[i]==c && pd[bitmap->size_x - i -1]==c)
				++i;
			cx = i;

			if (yu < cy && cx != bitmap->size_x / 2)
				cy = yu;

			++yu;
			--yd;
		}
	} else if (bitmap->bytes_per_pixel == 2) {
		uint16 c = *(uint16*)bitmap_line(bitmap,0);
		while (yu < yd && cx) {
			unsigned i;
			uint16* pu = (uint16*)bitmap_line(bitmap,yu);
			uint16* pd = (uint16*)bitmap_line(bitmap,yd);

			i = 0;
			while (i < cx && pu[i]==c && pu[bitmap->size_x - i -1]==c)
				++i;
			cx = i;

			i = 0;
			while (i < cx && pd[i]==c && pd[bitmap->size_x - i -1]==c)
				++i;
			cx = i;

			if (yu < cy && cx != bitmap->size_x / 2)
				cy = yu;

			++yu;
			--yd;
		}
	} else {
		cx = 0;
		cy = 0;
	}

	_cx = cx;
	_cy = cy;
}

// Update the backdrop image
static void text_backdrop_update(struct backdrop_data* back) {
	if (back->backdrop_path[0]) {
		if (!back->backdrop) {
			if (access(back->backdrop_path,F_OK | R_OK)!=0) {
				return;
			} else {
				video_color rgb[256];
				unsigned rgb_max;
				struct bitmap* bitmap = backdrop_load(back->backdrop_path, rgb, &rgb_max, 0);
				if (!bitmap)
					return;

				if (video_index() == VIDEO_FLAGS_INDEX_PACKED && bitmap->bytes_per_pixel != 1) {
					rgb_max = 256 - PALETTE_RESERVED_MAX;
					rgb_max = bitmap_reduction(back->backdrop_convert, rgb, rgb_max, bitmap, 0);
				}

				// set orientation (only flip)
				if (bitmap_orientation(bitmap,text_orientation & ORIENTATION_FLIP_XY, 0) != 0) {
					bitmap_free(bitmap);
					return;
				}

				bitmap_free(back->backdrop);
				back->backdrop = bitmap;

				unsigned cx = 0;
				unsigned cy = 0;
				text_backdrop_cut(bitmap, cx, cy);

				// set orientation (only mirror) and scaling
				for(unsigned i=0;i<video_size_x();++i) {
					unsigned x = cx + i * (back->backdrop->size_x - cx*2)/ back->real_backdrop_dx;
					if (text_orientation & ORIENTATION_MIRROR_X)
						x = back->backdrop->size_x - x - 1;
					back->backdrop_x_map[i] = x;
				}
				for(unsigned i=0;i<video_size_y();++i) {
					unsigned y = cy + i * (back->backdrop->size_y - cy*2) / back->real_backdrop_dy;
					if (text_orientation & ORIENTATION_MIRROR_Y)
						y = back->backdrop->size_y - y - 1;
					back->backdrop_y_map[i] = y;
				}

				// palette
				if (back->backdrop->bytes_per_pixel == 1) {
					if (video_index() == VIDEO_FLAGS_INDEX_RGB) {
						for(unsigned i=0;i<rgb_max;++i)
							video_rgb_make(&back->backdrop_color_map[i],rgb[i].red,rgb[i].green,rgb[i].blue);
					} else if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
						palette_set(rgb,rgb_max,back->backdrop_index_map);
					}
				} else {
					if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
						unsigned i;
						palette_set(rgb,rgb_max,back->backdrop_index_map);
						for(i=0;i<BITMAP_INDEX_MAX;++i)
							back->backdrop_convert[i] = back->backdrop_index_map[back->backdrop_convert[i]];
					}
				}
			}
		}
	} else {
		// palette
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED)
			palette_set(0,0,0);
	}


	if (back->backdrop) {
		if (video_bytes_per_pixel() == 1) {
			if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
				if (back->backdrop->bytes_per_pixel==1)
					text_put8pack_backdrop8(back);
				else
					text_put8pack_backdrop24(back);
			} else {
				if (back->backdrop->bytes_per_pixel==1)
					text_put8rgb_backdrop8(back);
				else
				text_put8rgb_backdrop24(back);
			}
		} else {
			if (back->backdrop->bytes_per_pixel==1)
				text_put16rgb_backdrop8(back);
			else
				text_put16rgb_backdrop24(back);
		}
	} else {
		text_clear(back->backdrop_x,back->backdrop_y,back->backdrop_dx,back->backdrop_dy,backdrop_missing_color >> 4);
	}

	// outline box
	text_box(back->backdrop_x-1,back->backdrop_y-1,back->backdrop_dx+2,back->backdrop_dy+2,backdrop_missing_color);

	/*
	if (video_bytes_per_pixel() == 1) {
		video_rgb color;
		if (video_index() == VIDEO_FLAGS_INDEX_PACKED) {
			color = 0xF;
		} else {
			video_rgb_make(&color,0xFF,0xFF,0xFF);
		}
		if (back->highlight)
			text_put8_box(back,color);
	} else {
		video_rgb color;
		video_rgb_make(&color,0xFF,0xFF,0xFF);
		if (back->highlight)
			text_put16_box(back,color);
	}
	*/
}

void text_backdrop_update() {
	for(int i=0;i<backdrop_mac;++i)
		text_backdrop_update(BACKDROP + i);
}

void text_clear() {
	memset(video_buffer,0,video_buffer_size);
}

static void text_copy(bool wait_vsync) {
	if (video_unchained_active()) {
		for(unsigned plane=0;plane<4;++plane) {
			video_unchained_plane_set(plane);
			unsigned char* buffer = video_buffer + plane * video_bytes_per_pixel();
			unsigned buffer_line_size = video_size_x() * video_bytes_per_pixel();
			_farsetsel( video_write_selector() );
			for(unsigned y=0;y<video_size_y();++y) {
				unsigned dst_off = video_write_line(y);
				unsigned char* src_ptr = buffer;
				for(unsigned x=0;x<video_size_x()/4;++x) {
					_farnspokeb(dst_off,*src_ptr);
					++dst_off;
					src_ptr += 4;
				}
				buffer += buffer_line_size;
			}
		}
	} else {
		unsigned char* buffer = video_buffer;
		unsigned buffer_line_size = video_size_x() * video_bytes_per_pixel();
		for(unsigned y=0;y<video_size_y();++y) {
			movedata(_my_ds(), (unsigned)buffer, video_write_selector(), video_write_line(y), buffer_line_size );
			buffer += buffer_line_size;
		}
	}

	if (video_index() == VIDEO_FLAGS_INDEX_PACKED)
		video_palette_set(text_palette,0,256,wait_vsync);
}

#ifdef __SNAPSHOT__
void text_snapshot() {
	static unsigned snapshot_num = 0;
	char snapshot_name[32];
	sprintf(snapshot_name,"snap%d.bmp",snapshot_num);
	snapshot_num++;

	video_snapshot_save(snapshot_name,0,0);
}
#endif

//---------------------------------------------------------------------------
// Key

struct key_cvt {
	const char* name;
	unsigned event;
	unsigned key0;
	unsigned key1;
};

static struct key_cvt KEYTAB[] = {
{"up",TEXT_KEY_UP, KEY_UP, KEY_MAX},
{"down",TEXT_KEY_DOWN, KEY_DOWN, KEY_MAX},
{"left",TEXT_KEY_LEFT, KEY_LEFT, KEY_MAX},
{"right",TEXT_KEY_RIGHT, KEY_RIGHT, KEY_MAX},
{"enter",TEXT_KEY_ENTER, KEY_ENTER, KEY_MAX},
{"esc",TEXT_KEY_ESC, KEY_ESC, KEY_MAX},
{"space",TEXT_KEY_SPACE, KEY_SPACE, KEY_MAX},
{"mode",TEXT_KEY_MODE, KEY_TAB, KEY_MAX},
{"home",TEXT_KEY_HOME, KEY_HOME, KEY_MAX},
{"end",TEXT_KEY_END, KEY_END, KEY_MAX},
{"pgup",TEXT_KEY_PGUP, KEY_PGUP, KEY_MAX},
{"pgdn",TEXT_KEY_PGDN, KEY_PGDN, KEY_MAX},
{"help",TEXT_KEY_HELP, KEY_F1, KEY_MAX},
{"group",TEXT_KEY_GROUP, KEY_F2, KEY_MAX},
{"type",TEXT_KEY_TYPE, KEY_F3, KEY_MAX},
{"exclude",TEXT_KEY_EXCLUDE, KEY_F4, KEY_MAX},
{"sort",TEXT_KEY_SORT, KEY_F5, KEY_MAX},
{"setgroup",TEXT_KEY_SETGROUP, KEY_F9, KEY_MAX},
{"settype",TEXT_KEY_SETTYPE, KEY_F10, KEY_MAX},
{"runclone",TEXT_KEY_RUN_CLONE, KEY_F12, KEY_MAX},
#ifdef __SNAPSHOT__
{"snapshot",TEXT_KEY_SNAPSHOT, KEY_DEL_PAD, KEY_MAX},
#endif
{"del",TEXT_KEY_DEL, KEY_DEL, KEY_MAX},
{"ins",TEXT_KEY_INS, KEY_INSERT, KEY_MAX},
{0,0,0,0}
};

static int keyboard_raw_poll() {
	poll_keyboard();

	for(const struct key_cvt* i=KEYTAB;i->name;++i)
		if ((i->key0 < KEY_MAX && key[i->key0])
			|| (i->key1 < KEY_MAX && key[i->key1]))
			return i->event;

	for(unsigned i=KEY_A;i<=KEY_Z;++i)
		if (key[i])
			return 'a' + i - KEY_A;

	return TEXT_KEY_NONE;
}

/**************************************************************************/

// Number of event repeat for second
#define KEY_repeat_REPEAT_0 2
#define KEY_repeat_REPEAT_1 20

static int key_poll() {
	static int key_repeat_last = TEXT_KEY_NONE;
	static clock_t key_repeat_last_time = 0;
	static bool key_repeat_last_counter = 0;

	int r = keyboard_raw_poll();
	if (r == TEXT_KEY_NONE)
		r = joystick_raw_poll();

	if (r != key_repeat_last || r == TEXT_KEY_NONE) {
		key_repeat_last = r;
		key_repeat_last_counter = 0;
		key_repeat_last_time = clock();
		return r;
	} else {
		if ((key_repeat_last_counter == 0 && (clock() - key_repeat_last_time > CLOCKS_PER_SEC / KEY_repeat_REPEAT_0)) ||
			(key_repeat_last_counter > 0 && (clock() - key_repeat_last_time > CLOCKS_PER_SEC / KEY_repeat_REPEAT_1))) {
			key_repeat_last_time = clock();
			++key_repeat_last_counter;
			return r;
		} else {
			return TEXT_KEY_NONE;
		}
	}
}

static int key_saved = TEXT_KEY_NONE;
static int key_last = TEXT_KEY_NONE;

int text_keypressed() {
	static clock_t key_pressed_last_time = 0;

	if (key_saved != TEXT_KEY_NONE)
		return 1;

	clock_t now = clock();

	/* don't check too fast */
	if (now - key_pressed_last_time >= CLOCKS_PER_SEC / 25) {
		key_pressed_last_time = now;

		key_saved = key_poll();

		if (key_saved != TEXT_KEY_NONE)
			return 1;
	}

	return 0;
}

void text_idle_repeat_reset() {
	key_last = TEXT_KEY_NONE;
}

void text_idle_time_reset() {
	text_idle_clock = clock();
}

static void text_idle() {
	if (key_last == TEXT_KEY_IDLE_0 && text_idle_0_rep && clock() - text_idle_clock > text_idle_0_rep * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_0;
	if (key_last == TEXT_KEY_IDLE_1 && text_idle_1_rep && clock() - text_idle_clock > text_idle_1_rep * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_1;
	if (text_idle_0 && clock() - text_idle_clock > text_idle_0 * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_0;
	if (text_idle_1 && clock() - text_idle_clock > text_idle_1 * CLOCKS_PER_SEC)
		key_saved = TEXT_KEY_IDLE_1;
}

unsigned text_getkey() {
	text_copy(false);

	while (!text_keypressed()) {
		text_idle();
	}

	text_idle_clock = clock();

	assert( key_saved != TEXT_KEY_NONE);

	key_last = key_saved;
	key_saved = TEXT_KEY_NONE;

#ifdef __SNAPSHOT__
	if (key_last == TEXT_KEY_SNAPSHOT)
		text_snapshot();
#endif

	return key_last;
}


static void text_key_insert(unsigned event, unsigned key0, unsigned key1) {
	unsigned i;
	for(i=0;KEYTAB[i].name && KEYTAB[i].event != event;++i);
	if (KEYTAB[i].name) {
		KEYTAB[i].key0 = key0;
		KEYTAB[i].key1 = key1;
	}
}

static unsigned text_string2event(const string& s) {
	unsigned i;
	for(i=0;KEYTAB[i].name && KEYTAB[i].name != s;++i);
	if (KEYTAB[i].name)
		return KEYTAB[i].event;
	else
		return TEXT_KEY_NONE;
}

bool text_key_in(const string& s) {
	string sevent;
	string skey0;
	string skey1;
	unsigned i = 0;

	while (i < s.length() && !isspace(s[i])) {
		sevent += s[i];
		++i;
	}
	
	while (i < s.length() && isspace(s[i]))
		++i;

	while (i < s.length() && !isspace(s[i])) {
		skey0 += s[i];
		++i;
	}
	
	while (i < s.length() && isspace(s[i]))
		++i;

	while (i < s.length() && !isspace(s[i])) {
		skey1 += s[i];
		++i;
	}
	
	while (i < s.length() && isspace(s[i]))
		++i;

	unsigned event = text_string2event(sevent);
	if (event == TEXT_KEY_NONE)
		return false;

	unsigned key0 = atoi(skey0.c_str());
	if (!key0)
		key0 = KEY_MAX;
	unsigned key1 = atoi(skey1.c_str());
	if (!key1)
		key1 = KEY_MAX;

	text_key_insert(event,key0,key1);

	if (i != s.length())
		return false;

	return true;
}

void text_key_out_all(ostream& os) {
	unsigned i;
	for(i=0;KEYTAB[i].name;++i) {
		os << "event " << KEYTAB[i].name << " ";
		if (KEYTAB[i].key0 < KEY_MAX)
			os << KEYTAB[i].key0;
		else
			os << "0";
		os << " ";
		if (KEYTAB[i].key1 < KEY_MAX)
			os << KEYTAB[i].key1;
		else
			os << "0";
		os << endl;
	}
}
